/** -------------------------------------------------
 *                   J-CHOCO
 *   Copyright (C) F. Laburthe, 1999-2003
 * --------------------------------------------------
 *    an open-source Constraint Programming Kernel
 *          for Research and Education
 * --------------------------------------------------
 *
 * file: choco.AbstractVar.java
 * last modified by Francois 28 aot 2003:14:59:09
 */
package choco;

import choco.prop.VarEvent;

import java.util.ArrayList;

/**
 * An abstract class for all implementations of domain variables
 */
public abstract class AbstractVar extends AbstractEntity implements Var {

  /**
   * A name may be associated to each variable
   */

  protected String name;

  /**
   * The variable var associated to this variable.
   */

  protected VarEvent event;


  /**
   * The list of constraints (listeners) observing the variable.
   */

  protected ArrayList constraints;


  /**
   * List of indices encoding the constraint network:
   * <i>v.indices[i]=j</i> means that v is the j-th variable
   * of its i-th constraint.
   */

  protected ArrayList indices;


  /**
   * Initializes a new variable.
   */

  public AbstractVar() {
    constraints = new ArrayList();
    indices = new ArrayList();
  }

  /**
   * useful for debugging
   */
  public String toString() {
    return name;
  }

  /**
   * Returns the variable var.
   */

  public VarEvent getEvent() {
    return event;
  }


  /**
   * Retrieve the i-th constraint involving the variable.
   */

  public Constraint getConstraint(int i) {
    return ((Constraint) constraints.get(i));
  }


  /**
   * Returns the number of constraints involving the variable.
   */

  public int getNbConstraints() {
    return constraints.size();
  }


  /**
   * Gets an array with all the constraints of the variable.
   */

  public Constraint[] getConstraints() {
    Constraint[] tab = new Constraint[constraints.size()];
    return (Constraint[]) constraints.toArray(tab);
  }


  /**
   * Gets an array with all the indices.
   */

  public Integer[] getVarIndices() {
    Integer[] tab = new Integer[constraints.size()];
    return (Integer[]) indices.toArray(tab);
  }

  /**
   * returns the index of the variable in it i-th constraint
   * @param constraintIndex the index of the constraint (among all constraints linked to the variable)
   * @return the index of the variable (0 if this is the first variable of that constraint)
   */
  public int getVarIndex(int constraintIndex) {
    return ((Integer) indices.get(constraintIndex)).intValue();
  }

  /**
   * Checks if the var associated with the <code>idx</code>th constraint
   * is active.
   */

  public boolean isActive(int idx1) {
    return this.event.isActive(idx1);
  }

  /**
   * removes (permanently) a constraint from the list of constraints connected to the variable
   * @param constraintIdx
   */
  public void eraseConstraint(int constraintIdx) {
    constraints.set(constraintIdx, null);
  }

  // ============================================
  // Managing Listeners.
  // ============================================

  /**
   * Adds a new constraints, and makes it active if needed.
   */
  // TODO: this needs to become backtrackable
  public int addConstraint(Constraint c, int varIdx, boolean active) {
    constraints.add(c);
    indices.add(new Integer(varIdx));

    int constraintIdx = constraints.size() - 1;

    this.event.addListener(constraintIdx, active);

    return constraintIdx;
  }

  // Si besoin de plus de precision, on pourra faire des methodes dediees
  // a differents types de Listener : I_IncInfListener....
  // soit addNewIncInfListener, etc. Chaque contrainte pourra alors choisir
  // entre appeler la methode par defaut qui modifie tout le monde pareil
  // (il faudra juste la modifier pour lexecuter sur chaque evenement) ou
  // appeler la methode specialisee...


  /**
   * Makes a listener passive.
   */

  public void deactivateConstraint(int constraintIdx) {
    this.event.deactivateListener(constraintIdx);
  }


  /**
   * Makes a listener active.
   */

  public void activateConstraint(int constraintIdx) {
    this.event.activateListener(constraintIdx);
  }
}
