/* --------------------------------------------------
 *                   CHOCO
 *   Copyright (C) F. Laburthe, 1999-2004
 * --------------------------------------------------
 *    an open-source Constraint Programming Kernel
 *          for Research and Education
 * --------------------------------------------------
 *
 * file: choco.Problem.java
 * last modified by Francois 28 aot 2003:14:59:09
 */
package choco;

import choco.integer.var.IntDomainVar;
import choco.integer.IntExp;
import choco.integer.var.IntTerm;
import choco.integer.IntVar;
import choco.integer.IntConstraint;
import choco.integer.constraints.*;
import choco.mem.Environment;
import choco.prop.*;
import choco.real.RealVar;
import choco.real.var.RealVarImpl;
import choco.bool.*;
import choco.set.SetVar;
import choco.set.var.SetVarImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Arrays;
import java.util.Comparator;
import java.util.logging.Logger;
import java.lang.reflect.Array;

/**
 * A problem is a global structure containing variables bound by listeners
 * as well as solutions or solver parameters
 */
public class Problem extends AbstractEntity implements IntModeler, BoolModeler {

  /**
   * a flag that may be turned on when debugging the Choco library
   */
  public static final boolean DEBUG = false;

  /**
   * All the constraints of the problem.
   */
  protected ArrayList constraints;

  /**
   * A constant denoting the true constraint (always satisfied)
   */
  public final Constraint TRUE = new ConstantConstraint(true);

  /**
   * A constant denoting the false constraint (never satisfied)
   */
  public final Constraint FALSE  = new ConstantConstraint(false);

  /**
   * A constant tool for comparing constraints
   */
  public final ConstraintComparator constraintComparator = new ConstraintComparator(this);

  /**
   * A constant tool for comparing integer variables
   */
  public final IntVarComparator intVarComparator = new IntVarComparator(this);

  /**
   * All the search intVars in the problem.
   */

  protected ArrayList intVars;

  /**
   * All the set intVars in the problem.
   */
  protected ArrayList setVars;

  /**
   * All the float vars in the problem.
   */
  protected ArrayList floatVars;

  /**
   * Precision of the search for a real problem.
   */
  protected double precision = 1.0e-6;

  /**
   * Minimal width reduction between two propagations.
   */
  protected double reduction = 0.99;

  /**
   * Allows to know if the problem is feasible (null if it was not solved).
   */
  public Boolean feasible = null;

  /**
   * True if the problem was solved.
   */
  protected boolean solved = false;

  /**
   * The variable modelling the objective function
   */
  protected IntDomainVar objective;

  /**
   *  Maximization / Minimization problem
   */
  protected boolean doMaximize;

  /**
   * The propagation engine to propagate during solving.
   */
  protected AbstractPropagationEngine propagationEngine;

  /**
   * The object controlling the global search exploration
   */
  protected Solver solver;

  /**
   * The environment managing the backtrackable data.
   */
  protected Environment environment;

  /**
   * Reference to the root Logger, an object for logging global trace statements (using the java.util.logging package)
   */
  private static Logger logger = Logger.getLogger("choco");

  /**
   * Constructs a problem.
   * @param env the Environment responsible for all the memory management
   */

  public Problem(Environment env) {
    this.environment = env;
    this.intVars = new ArrayList();
    this.setVars = new ArrayList();
    this.floatVars = new ArrayList();
    this.constraints = new ArrayList();
    this.propagationEngine = new ChocEngine(this);
    this.solver = new Solver(this);
  }

  /**
   * Constructs a problem.
   */

  public Problem() {
    this( new Environment() );
  }

  /**
   * Returns the memory environment used by the problem.
   */

  public final Environment getEnvironment() {
    return environment;
  }


  /** <i>Network management:</i>
   *    Retrieve a variable by its index (all search variables of
   *    the problem are numbered in sequence from 0 on)
   *    @param i   index of the variable in the problem
   */

  public final IntVar getIntVar(int i) {
    return (IntVar) intVars.get(i);
  }

  public int getIntVarIndex(IntDomainVar c) {
    return intVars.indexOf(c);
  }

  /**
   * retrieving the total number of variables
   * @return  the total number of variables in the problem
   */
  public final int getNbIntVars() {
    return intVars.size();
  }

  /**
   * retrieving the total number of constraints over integers
   * @return  the total number of constraints over integers in the problem
   */
  public final int getNbIntConstraints() {
    return constraints.size();
  }

  /** <i>Network management:</i>
   *    Retrieve a variable by its index (all search variables of
   *    the problem are numbered in sequence from 0 on)
   *    @param i   index of the variable in the problem
   */

  public final IntConstraint getIntConstraint(int i) {
    return (IntConstraint) constraints.get(i);
  }

  /**
   *    Returns the propagation engine associated to the problem
   */

  public PropagationEngine getPropagationEngine() {
    return getChocEngine();
  }

  /**
   * accessor returning the object controlling the propagation mechanism
   * @return the propagation engine
   */
  public final ChocEngine getChocEngine() {
    return (ChocEngine) propagationEngine;
  }

  /**
   * accessor returning the handler of the object responsible for the resolution of the problem.
   * It may be either a local search solver or a global search solver
   * @return  the object responsible for solving the problem
   */
  public final Solver getSolver() {
    return solver;
  }

  /** <i>Network management:</i>
   *    adding a constraint to the problem. Note that this does not
   *    propagate anything !
   *    @param c   the constraint to add
   */

  public void post(Constraint c) {
    constraints.add(c);
    c.addListener();
    ConstraintEvent event = (ConstraintEvent) c.getEvent();
    ChocEngine pe = getChocEngine();
    pe.registerEvent(event);
    pe.postConstAwake(c, true);
  }

  /** <i>Propagation:</i>
   *    Computes consistency on the problem (the problem may no longer
   *    be consistent since the last propagation because of listeners
   *    that have been posted and variables that have been reduced
   * @throws choco.ContradictionException
   */

  public void propagate() throws ContradictionException {
    ChocEngine pe = getChocEngine();
    boolean someEvents = true;
    while (someEvents) {
      EventQueue q = pe.getNextActiveEventQueue();
      if (q != null) {
        q.propagateSomeEvents();
      } else
        someEvents = false;
    }
    pe.checkCleanState();
  }

  /**
   * pushing one world on the stack
   */
  public final void worldPush() {
    environment.worldPush();
  }

  /**
   * pushing one world on the stack
   */
  public final void worldPop() {
    environment.worldPop();
  }

  /**
   * Backtracks to a given level in the search tree.
   */
  public final void worldPopUntil(int n) {
    while (environment.currentWorld > n) {
      worldPop();
    }
  }

  /**
   * returning the index of the current worl
   */
  public final int getWorldIndex() {
    return environment.getWorldIndex();
  }

  /** <i>Resolution:</i>
   *    Searches for the first solution. Either stops in the state of
   *    that solution or returns to the initial state when no solutions exist.
   * @param all Specifies wether all the solutions should be searched.
   */

  public Boolean solve(boolean all) {
    solver.firstSolution = !all;
    solver.generateSearchSolver(this);
//    solver.getSearchSolver().incrementalRun();
    solver.launch();
    return isFeasible();
  }

  public Boolean solve() {
    solver.firstSolution = true;
    solver.generateSearchSolver(this);
//    solver.getSearchSolver().incrementalRun();
    solver.launch();
    return isFeasible();
  }

  public boolean solveAll(){
    solver.firstSolution = false;
    solver.generateSearchSolver(this);
//    solver.getSearchSolver().incrementalRun();
    solver.launch();
    return true;
  }


  public Boolean nextSolution() {
    return solver.getSearchSolver().nextSolution();
  }

  /** <i>Resolution:</i>
   *    Searches for the solution minimizing the objective criterion.
   *    @param obj The variable modelling the optimization criterion
   *    @param restart If true, then a new search is restarted from scratch
   *                         after each solution is found;
   *                   otherwise a single branch-and-bound search is performed
   */

  public int minimize(IntVar obj, boolean restart) {
    return 1;
  }

  /** <i>resolution:</i>
   *    Searches for the solution maximizing the objective criterion.
   *    @param obj The variable modelling the optimization criterion
   *    @param restart If true, then a new search is restarted from scratch
   *                         after each solution is found;
   *                   otherwise a single branch-and-bound search is performed
   */
  public int maximize(IntVar obj, boolean restart) {
    return 1;
  }

  /**
   * As all Entity, must be able to return a Problem, returns itself.
   */

  public final Problem getProblem() {
    return this;
  }

  /** Creates a new search variable with an enumerated domain
   * @param name the name of the variable
   * @param min minimal allowed value (included in the domain)
   * @param max maximal allowed value (included in the domain)
   * @return the variable
   */
  public IntVar makeEnumIntVar(String name, int min, int max) {
    IntDomainVar v = new IntDomainVar(this, name, IntDomainVar.LIST, min, max);
    intVars.add(v);
    return v;
  }

  /** Creates a new search variable with an interval domain
   * @param name the name of the variable
   * @param min minimal allowed value (included in the domain)
   * @param max maximal allowed value (included in the domain)
   * @return the variable
   */
  public IntVar makeBoundIntVar(String name, int min, int max) {
    IntDomainVar v = new IntDomainVar(this, name, IntDomainVar.BOUNDS, min, max);
    intVars.add(v);
    return v;
  }

  // TODO: can be optimized (no need to attach propagation events)
  public IntVar makeConstantIntVar(String name, int val) {
    return makeBoundIntVar(name,val,val);
  }

  public IntVar makeConstantIntVar(int val) {
    return makeConstantIntVar("",val);
  }

  /** Creates a simple linear term from one coefficient and one variable
   * @param a the coefficient
   * @param x the variable
   * @return the term
   */
  public IntExp mult(int a, IntExp x) {
    if (a != 0) {
      IntTerm t = new IntTerm(1);
      t.setCoefficient(0, a);
      t.setVariable(0, (IntDomainVar) x);
      return t;
    } else {
      return null;
    }
  }

  /**
   * Utility method for constructing a term from two lists of variables, list of coeffcicients and constants
   * @param coeffs1 coefficients from the first term
   * @param vars1   variables from the first term
   * @param cste1   constant from the fisrt term
   * @param coeffs2 coefficients from the second term
   * @param vars2   variables from the second term
   * @param cste2   constant from the second term
   * @return the term (a fresh one)
   */
  protected IntExp plus(
      int[] coeffs1, IntDomainVar[] vars1, int cste1, int[] coeffs2, IntDomainVar[] vars2, int cste2) {
    int n1 = vars1.length;
    int n2 = vars2.length;
    IntTerm t = new IntTerm(n1 + n2);
    for (int i = 0; i < n1; i++) {
      t.setVariable(i, vars1[i]);
      t.setCoefficient(i, coeffs1[i]);
    }
    for (int i = 0; i < n2; i++) {
      t.setVariable(n1 + i, vars2[i]);
      t.setCoefficient(n1 + i, coeffs2[i]);
    }
    t.setConstant(cste1 + cste2);
    return t;
  }

  /** Adding two terms one to another
   * @param t1 first term
   * @param t2 second term
   * @return the term (a fresh one)
   */
  public IntExp plus(IntExp t1, IntExp t2) {
    if (t1 instanceof IntTerm) {
      if (t2 instanceof IntTerm) {
        return plus(
            ((IntTerm) t1).getCoefficients(),
            ((IntTerm) t1).getVariables(),
            ((IntTerm) t1).getConstant(),
            ((IntTerm) t2).getCoefficients(),
            ((IntTerm) t2).getVariables(),
            ((IntTerm) t2).getConstant());
      } else if (t2 instanceof IntDomainVar) {
        return plus(
            ((IntTerm) t1).getCoefficients(),
            ((IntTerm) t1).getVariables(),
            ((IntTerm) t1).getConstant(),
            new int[]{1},
            new IntDomainVar[]{(IntDomainVar) t2},
            0);
      } else {
        throw new Error("IntExp not a term, not a var");
      }
    } else if (t1 instanceof IntDomainVar) {
      if (t2 instanceof IntTerm) {
        return plus(
            new int[]{1},
            new IntDomainVar[]{(IntDomainVar) t1},
            0,
            ((IntTerm) t2).getCoefficients(),
            ((IntTerm) t2).getVariables(),
            ((IntTerm) t2).getConstant());
      } else if (t2 instanceof IntDomainVar) {
        IntTerm t = new IntTerm(2);
        t.setCoefficient(0, 1);
        t.setCoefficient(1, 1);
        t.setVariable(0, (IntDomainVar) t1);
        t.setVariable(1, (IntDomainVar) t2);
        t.setConstant(0);
        return t;
      } else
        throw new Error("IntExp not a term, not a var");
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public IntExp plus(IntExp t, int c) {
    if (t instanceof IntTerm) {
      IntTerm t2 = new IntTerm((IntTerm) t);
      t2.setConstant(((IntTerm) t).getConstant() + c);
      return t2;
    } else if (t instanceof IntDomainVar) {
      IntTerm t2 = new IntTerm(1);
      t2.setCoefficient(0,1);
      t2.setVariable(0, (IntDomainVar) t);
      t2.setConstant(c);
      return t2;
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public final IntExp plus(int c, IntExp t1) {
    return plus(t1,c);
  }

  /** Subtracting two terms one from another
   * @param t1 first term
   * @param t2 second term
   * @return the term (a fresh one)
   */
  public IntExp minus(IntExp t1, IntExp t2) {
    if (t1 instanceof IntTerm) {
      if (t2 instanceof IntTerm) {
        int[] coeffs2 = ((IntTerm) t2).getCoefficients();
        int n2 = coeffs2.length;
        int[] oppcoeffs2 = new int[n2];
        for (int i = 0; i < n2; i++) {
          oppcoeffs2[i] = -(coeffs2[i]);
        }
        return plus(
            ((IntTerm) t1).getCoefficients(),
            ((IntTerm) t1).getVariables(),
            ((IntTerm) t1).getConstant(),
            oppcoeffs2,
            ((IntTerm) t2).getVariables(),
            -((IntTerm) t2).getConstant() );
      } else if (t2 instanceof IntDomainVar) {
        return plus(
            ((IntTerm) t1).getCoefficients(),
            ((IntTerm) t1).getVariables(),
            ((IntTerm) t1).getConstant(),
            new int[]{-1},
            new IntDomainVar[]{(IntDomainVar) t2},
            0);
      } else {
        throw new Error("IntExp not a term, not a var");
      }
    } else if (t1 instanceof IntDomainVar) {
      if (t2 instanceof IntTerm) {
        int[] coeffs2 = ((IntTerm) t2).getCoefficients();
        int n2 = coeffs2.length;
        int[] oppcoeffs2 = new int[n2];
        for (int i = 0; i < n2; i++) {
          oppcoeffs2[i] = -(coeffs2[i]);
        }
        return plus(
            new int[]{1},
            new IntDomainVar[]{(IntDomainVar) t1},
            0,
            oppcoeffs2,
            ((IntTerm) t2).getVariables(),
            -((IntTerm) t2).getConstant());
      } else if (t2 instanceof IntDomainVar) {
        IntTerm t = new IntTerm(2);
        t.setCoefficient(0, 1);
        t.setCoefficient(1, -1);
        t.setVariable(0, (IntDomainVar) t1);
        t.setVariable(1, (IntDomainVar) t2);
        t.setConstant(0);
        return t;
      } else
        throw new Error("IntExp not a term, not a var");
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public IntExp minus(IntExp t, int c) {
    if (t instanceof IntTerm) {
      IntTerm t2 = new IntTerm((IntTerm) t);
      t2.setConstant(((IntTerm) t).getConstant() - c);
      return t2;
    } else if (t instanceof IntDomainVar) {
      IntTerm t2 = new IntTerm(1);
      t2.setCoefficient(0,1);
      t2.setVariable(0, (IntDomainVar) t);
      t2.setConstant(-c);
      return t2;
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public IntExp minus(int c, IntExp t) {
    if (t instanceof IntTerm) {
      IntTerm t1 = (IntTerm) t;
      int n = t1.getSize();
      IntTerm t2 = new IntTerm(n);
      for (int i=0; i<n; i++) {
        t2.setCoefficient(i,-t1.getCoefficient(i));
        t2.setVariable(i, t1.getVariable(i));
      }
      t2.setConstant(c - t1.getConstant());
      return t2;
    } else if (t instanceof IntDomainVar) {
      IntTerm t2 = new IntTerm(1);
      t2.setCoefficient(0,-1);
      t2.setVariable(0, (IntDomainVar) t);
      t2.setConstant(c);
      return t2;
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  /** Building a term from a scalar product of coefficients and variables
   * @param lc the array of coefficients
   * @param lv the array of variables
   * @return the term
   */
  public IntExp scalar(int[] lc, IntVar[] lv) {
    int n = lc.length;
    assert(lv.length == n);
    IntTerm t = new IntTerm(n);
    for (int i = 0; i < n; i++) {
      t.setCoefficient(i, lc[i]);
      if (lv[i] instanceof IntDomainVar) {
        t.setVariable(i, (IntDomainVar) lv[i]);
      } else {
        throw new Error("unknown kind of IntVar");
      }
    }
    return t;
  }

  /**
   * Building a term from a scalar product of coefficients and variables
   *
   * @param lc the array of coefficients
   * @param lv the array of variables
   * @return the term
   */
  public final IntExp scalar(IntVar[] lv, int[] lc) {
    return scalar(lc,lv);
  }

  /**
   * Building a term from a sum of integer expressions
   *
   * @param lv the array of integer expressions
   * @return the term
   */
  public IntExp sum(IntExp[] lv) {
    int n = lv.length;
    IntTerm t = new IntTerm(n);
    for (int i = 0; i < n; i++) {
      t.setCoefficient(i, 1);
      if (lv[i] instanceof IntDomainVar) {
        t.setVariable(i, (IntDomainVar) lv[i]);
      } else {
        throw new Error("unexpected kind of IntExp");
      }
    }
    return t;
  }

  /** Creates a constraint by stating that a term is not equal than a constant
   * @param x the expression
   * @param c the constant
   * @return the linear disequality constraint
   */
  public Constraint neq(IntExp x, int c) {
    if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        return new NotEqualXYC(t.getVariable(0), t.getVariable(1), (c - t.getConstant()) / t.getCoefficient(0));
      } else {
        return new IntLinComb(((IntTerm) x).getVariables(), ((IntTerm) x).getCoefficients(), -(c), IntLinComb.NEQ);
      }
    } else if (x instanceof IntDomainVar) {
      return new NotEqualXC((IntDomainVar) x, c);
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public final Constraint neq(int c, IntExp x) {
    return neq(x, c);
  }

  public Constraint neq(IntExp x, IntExp y) {
    if (x instanceof IntTerm) {
      return neq(minus(x,y),0);
    } else if (x instanceof IntDomainVar) {
      if (y instanceof IntTerm) {
        return neq(minus(x,y),0);
      } else if (y instanceof IntDomainVar) {
        return new NotEqualXYC((IntDomainVar) x, (IntDomainVar) y, 0);
      } else if (y == null) {
        return neq(x,0);
      } else {
        throw new Error("IntExp not a term, not a var");
      }
    } else if (x == null) {
      return neq(0,y);
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public Constraint geq(IntExp x, IntExp y) {
      if (x instanceof IntVar && y instanceof IntVar) {
        return new GreaterOrEqualXYC((IntVar) x, (IntVar) y, 0);
      } else if ((x instanceof IntTerm || x instanceof IntVar) && (y instanceof IntTerm || y instanceof IntVar)) {
        return geq(this.minus(x, y), 0);
      } else if (y == null) {
        return geq(x,0);
      } else if (x == null) {
        return leq(x,0);
      } else {
        throw new Error("IntExp not a good exp");
      }
  }

  /** Creates a constraint by stating that a term is greater or equal than a constant
   * @param x the expression
   * @param c the constant
   * @return the linear inequality constraint
   */
  public Constraint geq(IntExp x, int c) {
    if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        if (t.getCoefficient(0) > 0) {
          return new GreaterOrEqualXYC(t.getVariable(0), t.getVariable(1), (c - t.getConstant()) / t.getCoefficient(0));
        } else {
          return new GreaterOrEqualXYC(t.getVariable(1), t.getVariable(0),  (c - t.getConstant()) / t.getCoefficient(1));
        }
      } else {
        return new IntLinComb(((IntTerm) x).getVariables(), ((IntTerm) x).getCoefficients(), -(c), IntLinComb.GEQ);
      }
    } else if (x instanceof IntDomainVar) {
      return new GreaterOrEqualXC((IntDomainVar) x, c);
    } else if (x == null) {
      if (c <= 0) {
        return TRUE;
      } else {
        return FALSE;
      }
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public Constraint geq(int c, IntExp x) {
    if (x instanceof IntTerm) {
      int[] coeffs = ((IntTerm) x).getCoefficients();
      int n = coeffs.length;
      int[] oppcoeffs = new int[n];
      for (int i = 0; i < n; i++) {
        oppcoeffs[i] = -(coeffs[i]);
      }
      return new IntLinComb(((IntTerm) x).getVariables(), oppcoeffs, c, IntLinComb.GEQ);
    } else if (x instanceof IntDomainVar) {
      return new LessOrEqualXC((IntDomainVar) x, c);
    } else if (x == null) {
      if (c <= 0) {
        return TRUE;
      } else {
        return FALSE;
      }
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public final Constraint gt(IntExp x, IntExp y) {
    return geq(minus(x,y),1);
  }

  public final Constraint gt(IntExp x, int c) {
    return geq(x,c + 1);
  }

  public final Constraint gt(int c, IntExp x) {
    return geq(c - 1,x);
  }

  public Constraint eq(IntExp x, IntExp y) {
      if (x instanceof IntVar && y instanceof IntVar) {
        return new EqualXYC((IntVar) x, (IntVar) y, 0);
      } else if ((x instanceof IntTerm || x instanceof IntVar) && (y instanceof IntTerm || y instanceof IntVar)) {
        return eq(this.minus(x, y), 0);
      } else if (x == null) {
        return eq(0,y);
      } else if (y == null) {
        return eq(x,0);
      } else {
        throw new Error("IntExp not a good exp");
      }
  }

  public Constraint eq(IntExp x, int c) {
    if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        return new EqualXYC(t.getVariable(0), t.getVariable(1), (c - t.getConstant()) / t.getCoefficient(0));
      } else {
      return new IntLinComb(((IntTerm) x).getVariables(), ((IntTerm) x).getCoefficients(), -(c), IntLinComb.EQ);
      }
    } else if (x instanceof IntDomainVar) {
      return new EqualXC((IntDomainVar) x, c);
    } else {
      throw new Error("IntExp not a term, not a var");
    }
  }

  public final Constraint eq(int c, IntExp x) {
    return eq(x, c);
  }

  public final Constraint leq(IntExp x, int c) {
    return geq(c, x);
  }

  public final Constraint leq(int c, IntExp x) {
    return geq(x, c);
  }

  public final Constraint leq(IntExp x, IntExp y) {
    return geq(y, x);
  }

  public final Constraint lt(IntExp x, int c) {
    return gt(c,x);
  }

  public final Constraint lt(int c, IntExp x) {
    return gt(x,c);
  }

  public final Constraint lt(IntExp x, IntExp y) {
    return gt(y,x);
  }

  /**
   * test if the problem has been found to be feasible (there exist solutions) or not.
   * @return Boolean.TRUE if a solution was found,
   *         Boolean.FALSE if the problem was proven infeasible,
   *         null otherwise
   */
  public final Boolean isFeasible() {
    return feasible;
  }

  public final void setMinimizationObjective(IntVar obj) {
    objective = (IntDomainVar) obj;
    doMaximize = false;
  }

  public final void setMaximizationObjective(IntVar obj) {
    objective = (IntDomainVar) obj;
    doMaximize = true;
  }

  public final boolean isConsistent() {
    int nbc = getNbIntConstraints();
    boolean consistent = true;
    for (int i=0; (i<nbc && consistent); i++) {
      IntConstraint c = getIntConstraint(i);
      consistent = (consistent && c.isConsistent());
    }
    return consistent;
  }

  public String pretty() {
    StringBuffer buf = new StringBuffer("Pb[" + (getNbIntVars() + getNbRealVars() + getNbSetVars()) + " vars, " + getNbIntConstraints() + " cons]\n");
    for (int i=0; i<getNbIntVars(); i++) {
      buf.append(getIntVar(i).pretty());
      buf.append("\n");
    }
    for (int i = 0; i < floatVars.size(); i++) {
      RealVar realVar = (RealVar) floatVars.get(i);
      buf.append(realVar);
    }
    for (int i = 0; i < setVars.size(); i++) {
      buf.append(getSetVar(i).pretty());
      buf.append("\n");
    }
    return new String(buf);
  }

  /**
   * Creates a one dimensional array of integer variables
   *
   * @param name the name of the array (a prefix shared by all individual IntVars)
   * @param dim  the number of entries
   * @param min  the minimal domain value for all variables in the array
   * @param max  the maximal domain value for all variables in the array
   */
  public IntVar[] makeBoundIntVarArray(String name, int dim, int min, int max) {
    IntVar[] res = new IntVar[dim];
    for (int i=0; i<dim; i++) {
      res[i] = makeBoundIntVar(name + "[" + String.valueOf(i) + "]", min, max);
    }
    return res;
  }

  /**
   * Creates a one dimensional array of integer variables
   *
   * @param name the name of the array (a prefix shared by all individual IntVars)
   * @param dim1 the number of entries for the first index
   * @param dim2 the number of entries for the second index
   * @param min  the minimal domain value for all variables in the array
   * @param max  the maximal domain value for all variables in the array
   */
  public IntVar[][] makeBoundIntVarArray(String name, int dim1, int dim2, int min, int max) {
    IntVar[][] res = new IntVar[dim1][dim2];
    for (int i=0; i<dim1; i++) {
      for (int j=0; j<dim2; j++) {
        res[i][j] = makeBoundIntVar(name + "[" + String.valueOf(i) + ", " + String.valueOf(j) + "]", min, max);
      }
    }
    return res;
  }

  /**
   * removes (permanently) a constraint from the overall list of constraints of the problem
   * @param constraintIdx
   */
  public void eraseConstraint(int constraintIdx) {
    constraints.set(constraintIdx, null);
  }

  public int getConstraintIndex(AbstractConstraint c) {
    return constraints.indexOf(c);
  }

  protected class ConstraintComparator implements Comparator {
    /**
     *  each problem has its own constraint comparator
     */
    private Problem problem;

    protected ConstraintComparator(Problem pb) {
      problem = pb;
    }

    public int compare(Object o1, Object o2) {
      int idx1 = problem.getConstraintIndex((AbstractConstraint) o1);
      int idx2 = problem.getConstraintIndex((AbstractConstraint) o2);
      if ((idx1 == -1) || (idx2 == -1)) {
        throw new UnsupportedOperationException("unable to compare constraints outside of their common Problem");
      } else {
        return (idx1 - idx2);
      }
    }
  }

   public void sortByConstraintIndex(AbstractConstraint[] constraintList) {
    Arrays.sort(constraintList, constraintComparator);
  }

   protected class IntVarComparator implements Comparator {
    /**
     *  each problem has its own int var comparator
     */
    private Problem problem;

    protected IntVarComparator(Problem pb) {
      problem = pb;
    }

    public int compare(Object o1, Object o2) {
      int idx1 = problem.getIntVarIndex((IntDomainVar) o1);
      int idx2 = problem.getIntVarIndex((IntDomainVar) o2);
      if ((idx1 == -1) || (idx2 == -1)) {
        throw new UnsupportedOperationException("unable to compare integer variables outside of their common Problem");
      } else {
        return (idx1 - idx2);
      }
    }
  }

   public void sortByIntVarIndex(IntVar[] intVarList) {
    Arrays.sort(intVarList, intVarComparator);
  }

  /**
   * Returns a real variable.
   * @param i index of the variable
   * @return the i-th real variable
   */
  public final RealVar getRealVar(int i) {
    return (RealVar) floatVars.get(i);
  }

  /**
   * Returns the number of variables modelling real numbers.
   */
  public final int getNbRealVars() {
    return floatVars.size();
  }

  /**
   * Creates a real variable.
   * @param name
   * @param min
   * @param max
   * @return the new object (RealVar)
   */
  public RealVar makeRealVar(String name, double min, double max) {
    RealVar v = new RealVarImpl(this, name, min, max);
    floatVars.add(v);
    return v;
  }

  public double getPrecision() {
    return precision;
  }

  public void setPrecision(double precision) {
    this.precision = precision;
  }

  public double getReduction() {
    return reduction;
  }

  public void setReduction(double reduction) {
    this.reduction = reduction;
  }

  /**
   * Returns a set variable.
   * @param i index of the variable
   * @return the i-th real variable
   */
  public final SetVar getSetVar(int i) {
    return (SetVar) setVars.get(i);
  }

  /**
   * Returns the number of variables modelling real numbers.
   */
  public final int getNbSetVars() {
    return setVars.size();
  }

  /**
   * Creates a set variable.
   * @param name
   * @param a : the first value of the initial enveloppe
   * @param b : the last value of the initial enveloppe
   * @return the new object (RealVar)
   */
  public SetVar makeSetVar(String name, int a, int b) {
    SetVar v = new SetVarImpl(this, name, a, b);
    setVars.add(v);
    return v;
  }

  public Constraint makeDisjunction(Constraint[] alternatives) {
    if (alternatives.length == 0) {
      throw new UnsupportedOperationException();
    } else if (alternatives.length == 1) {
      return alternatives[0];
    } else if (alternatives.length == 2) {
      return new BinDisjunction((AbstractConstraint) alternatives[0], (AbstractConstraint) alternatives[1]);
    } else {
      return new LargeDisjunction(alternatives);
     }
  }

  public final Constraint or(Constraint[] constList) {
    return makeDisjunction(constList);
  }

  public final Constraint or(Constraint c0, Constraint c1, Constraint c2) {
    return makeDisjunction(new Constraint[]{c0,c1,c2});
  }

  public final Constraint or(Constraint c0, Constraint c1, Constraint c2, Constraint c3) {
    return makeDisjunction(new Constraint[]{c0,c1,c2,c3});
  }

  public final Constraint or(Constraint c0, Constraint c1) {
    Constraint[] alternatives;
    if (c0 instanceof BinDisjunction) {
      if (c1 instanceof BinDisjunction) {
        alternatives = new Constraint[4];
        alternatives[0] = ((BinDisjunction) c0).getSubConstraint(0);
        alternatives[1] = ((BinDisjunction) c0).getSubConstraint(1);
        alternatives[2] = ((BinDisjunction) c1).getSubConstraint(0);
        alternatives[3] = ((BinDisjunction) c1).getSubConstraint(1);
      } else if (c1 instanceof LargeDisjunction) {
        int nb1 = ((LargeDisjunction) c1).getNbSubConstraints();
        alternatives = new Constraint[2 + nb1];
        alternatives[0] = ((BinDisjunction) c0).getSubConstraint(0);
        alternatives[1] = ((BinDisjunction) c0).getSubConstraint(1);
        for (int constIdx = 0; constIdx<nb1; constIdx++) {
          alternatives[2 + constIdx] = ((LargeDisjunction) c1).getSubConstraint(constIdx);
        }
      } else { // c1 is not a disjunction
        alternatives = new Constraint[3];
        alternatives[0] = ((BinDisjunction) c0).getSubConstraint(0);
        alternatives[1] = ((BinDisjunction) c0).getSubConstraint(1);
        alternatives[2] = c1;
      }
    } else if (c0 instanceof LargeDisjunction) {
      if (c1 instanceof BinDisjunction) {
        int nb0 = ((LargeDisjunction) c0).getNbSubConstraints();
        alternatives = new Constraint[2 + nb0];
        alternatives[0] = ((BinDisjunction) c1).getSubConstraint(0);
        alternatives[1] = ((BinDisjunction) c1).getSubConstraint(1);
        for (int constIdx = 0; constIdx<((LargeDisjunction) c0).getNbSubConstraints(); constIdx++) {
          alternatives[2 + constIdx] = ((LargeDisjunction) c0).getSubConstraint(constIdx);
        }
      } else if (c1 instanceof LargeDisjunction) {
        int nb0 = ((LargeDisjunction) c0).getNbSubConstraints();
        int nb1 = ((LargeDisjunction) c1).getNbSubConstraints();
        alternatives = new Constraint[nb0 + nb1];
        for (int constIdx = 0; constIdx<nb0; constIdx++) {
          alternatives[constIdx] = ((LargeDisjunction) c0).getSubConstraint(constIdx);
        }
        for (int constIdx = 0; constIdx<nb1; constIdx++) {
          alternatives[nb0 + constIdx] = ((LargeDisjunction) c1).getSubConstraint(constIdx);
        }
      } else {
        int nb0 = ((LargeDisjunction) c0).getNbSubConstraints();
        alternatives = new Constraint[nb0 + 1];
        alternatives[0] = ((BinDisjunction) c1).getSubConstraint(0);
        alternatives[1] = ((BinDisjunction) c1).getSubConstraint(1);
        for (int constIdx = 0; constIdx<nb0; constIdx++) {
          alternatives[constIdx] = ((LargeDisjunction) c0).getSubConstraint(constIdx);
        }
        alternatives[nb0] = c1;
      }
    } else { // c0 does not implement a disjunction
      if (c1 instanceof BinDisjunction) {
        alternatives = new Constraint[3];
        alternatives[0] = c0;
        alternatives[1] = ((BinDisjunction) c1).getSubConstraint(0);
        alternatives[2] = ((BinDisjunction) c1).getSubConstraint(1);
      } else if (c1 instanceof LargeDisjunction) {
        int nb1 = ((LargeDisjunction) c1).getNbSubConstraints();
        alternatives = new Constraint[1 + nb1];
        alternatives[0] = c0;
        for (int constIdx = 0; constIdx<nb1; constIdx++) {
          alternatives[1 + constIdx] = ((LargeDisjunction) c1).getSubConstraint(constIdx);
        }
      } else { // c1 does not implement a disjunction
        alternatives = new Constraint[2];
        alternatives[0] = c0;
        alternatives[1] = c1;
      }
    }
  return makeDisjunction(alternatives);
  }

  public Constraint makeConjunction(Constraint[] branches) {
    if (branches.length == 0) {
      throw new UnsupportedOperationException();
    } else if (branches.length == 1) {
     return branches[0];
    } else if (branches.length == 2) {
      return new BinConjunction((AbstractConstraint) branches[0], (AbstractConstraint) branches[1]);
    } else {
      return new LargeConjunction(branches);
    }
  }

  public final Constraint and(Constraint[] constList) {
    return makeConjunction(constList);
  }

  public final Constraint and(Constraint c0, Constraint c1, Constraint c2) {
    return makeConjunction(new Constraint[]{c0,c1,c2});
  }

  public final Constraint and(Constraint c0, Constraint c1, Constraint c2, Constraint c3) {
    return makeConjunction(new Constraint[]{c0,c1,c2,c3});
  }

  public final Constraint and(Constraint c0, Constraint c1) {
    Constraint[] branches;
    if (c0 instanceof BinConjunction) {
      if (c1 instanceof BinConjunction) {
        branches = new Constraint[4];
        branches[0] = ((BinConjunction) c0).getSubConstraint(0);
        branches[1] = ((BinConjunction) c0).getSubConstraint(1);
        branches[2] = ((BinConjunction) c1).getSubConstraint(0);
        branches[3] = ((BinConjunction) c1).getSubConstraint(1);
      } else if (c1 instanceof LargeConjunction) {
        int nb1 = ((LargeConjunction) c1).getNbSubConstraints();
        branches = new Constraint[2 + nb1];
        branches[0] = ((BinConjunction) c0).getSubConstraint(0);
        branches[1] = ((BinConjunction) c0).getSubConstraint(1);
        for (int constIdx = 0; constIdx<nb1; constIdx++) {
          branches[2 + constIdx] = ((LargeConjunction) c1).getSubConstraint(constIdx);
        }
      } else { // c1 is not a Conjunction
        branches = new Constraint[3];
        branches[0] = ((BinConjunction) c0).getSubConstraint(0);
        branches[1] = ((BinConjunction) c0).getSubConstraint(1);
        branches[2] = c1;
      }
    } else if (c0 instanceof LargeConjunction) {
      if (c1 instanceof BinConjunction) {
        int nb0 = ((LargeConjunction) c0).getNbSubConstraints();
        branches = new Constraint[2 + nb0];
        branches[0] = ((BinConjunction) c1).getSubConstraint(0);
        branches[1] = ((BinConjunction) c1).getSubConstraint(1);
        for (int constIdx = 0; constIdx<((LargeConjunction) c0).getNbSubConstraints(); constIdx++) {
          branches[2 + constIdx] = ((LargeConjunction) c0).getSubConstraint(constIdx);
        }
      } else if (c1 instanceof LargeConjunction) {
        int nb0 = ((LargeConjunction) c0).getNbSubConstraints();
        int nb1 = ((LargeConjunction) c1).getNbSubConstraints();
        branches = new Constraint[nb0 + nb1];
        for (int constIdx = 0; constIdx<nb0; constIdx++) {
          branches[constIdx] = ((LargeConjunction) c0).getSubConstraint(constIdx);
        }
        for (int constIdx = 0; constIdx<nb1; constIdx++) {
          branches[nb0 + constIdx] = ((LargeConjunction) c1).getSubConstraint(constIdx);
        }
      } else {
        int nb0 = ((LargeConjunction) c0).getNbSubConstraints();
        branches = new Constraint[nb0 + 1];
        branches[0] = ((BinConjunction) c1).getSubConstraint(0);
        branches[1] = ((BinConjunction) c1).getSubConstraint(1);
        for (int constIdx = 0; constIdx<nb0; constIdx++) {
          branches[constIdx] = ((LargeConjunction) c0).getSubConstraint(constIdx);
        }
        branches[nb0] = c1;
      }
    } else { // c0 does not implement a Conjunction
      if (c1 instanceof BinConjunction) {
        branches = new Constraint[3];
        branches[0] = c0;
        branches[1] = ((BinConjunction) c1).getSubConstraint(0);
        branches[2] = ((BinConjunction) c1).getSubConstraint(1);
      } else if (c1 instanceof LargeConjunction) {
        int nb1 = ((LargeConjunction) c1).getNbSubConstraints();
        branches = new Constraint[1 + nb1];
        branches[0] = c0;
        for (int constIdx = 0; constIdx<nb1; constIdx++) {
          branches[1 + constIdx] = ((LargeConjunction) c1).getSubConstraint(constIdx);
        }
      } else { // c1 does not implement a Conjunction
        branches = new Constraint[2];
        branches[0] = c0;
        branches[1] = c1;
      }
    }
  return makeConjunction(branches);
  }

  public final Constraint implies(Constraint c1, Constraint c2) {
    return or(not(c1), c2);
  }

  public Constraint ifThen(Constraint c1, Constraint c2) {
    return new Guard((AbstractConstraint) c1, (AbstractConstraint) c2);
  }

  public Constraint ifOnlyIf(Constraint c1, Constraint c2) {
    return new Equiv((AbstractConstraint) c1, (AbstractConstraint) c2);
  }

  public final Constraint not(Constraint c) {
    return c.opposite();
  }

 public Constraint atleast(Constraint[] constList, int nbTrueConstraints) {
    IntVar cardVar = makeConstantIntVar(nbTrueConstraints);
    return new Cardinality(constList, cardVar, true, false);
  }

  public Constraint atmost(Constraint[] constList, int nbTrueConstraints) {
    IntVar cardVar = makeConstantIntVar(nbTrueConstraints);
    return new Cardinality(constList, cardVar, false, true);
  }

  public Constraint card(Constraint[] constList, IntVar nbTrueConstraints) {
    return new Cardinality(constList, nbTrueConstraints, true, true);
  }
  
}
