package choco;

import choco.integer.var.IntDomainVar;
import choco.integer.search.IncreasingDomain;
import choco.integer.search.MinDomain;
import choco.integer.search.IValIterator;
import choco.search.*;

// *********************************************
// *                   J-CHOCO                 *
// *   Copyright (c) F. Laburthe, 1999-2003    *
// *********************************************
// * Event-base contraint programming Engine   *
// *********************************************

// CVS Information
// File:               $RCSfile: Solver.java,v $
// Version:            $Revision: 1.2 $
// Last Modification:  $Date: 2004/06/16 16:07:49 $
// Last Contributor:   $Author: hadrien $

/**
 * This class serves both as a factory and as a handler for AbstractGlobalSearchSolvers:
 */
public class Solver extends AbstractEntity {

  /**
   * The variable modelling the objective function
   */
  protected choco.integer.var.IntDomainVar objective;

  /**
   *  Maximization / Minimization problem
   */
  protected boolean doMaximize;

  /**
   * Do we want to restart a new search after each solution
   */
  protected boolean restart = false;

  /**
   * do we want to explore one or all solutions (default=one solution)
   */
  protected boolean firstSolution = true;

  /**
   * The object controlling the global search exploration
   */
  protected AbstractGlobalSearchSolver solver;

  /**
   * Variable selector
   */
  protected IVarSelector varSelector=null;

  /**
   * Value iterator
   */
  protected IValIterator valIterator=null;

  protected int timeLimit = Integer.MAX_VALUE;

  protected int nodeLimit = Integer.MAX_VALUE;

  public AbstractGlobalSearchSolver getSearchSolver() {
    return solver;
  }

  public Solver(Problem pb) {
    super(pb);
  }

  public void generateSearchSolver(Problem pb) {
    problem = pb;
    if (null == objective) {
      solver = new Solve(this.getProblem());
    } else if (restart) {
      solver = new OptimizeWithRestarts(objective, doMaximize);
    } else {
      solver = new BranchAndBound(objective, doMaximize);
    }
    solver.stopAtFirstSol = firstSolution;

    solver.limits.add(new TimeLimit(solver,timeLimit));
    solver.limits.add(new NodeLimit(solver,nodeLimit));

    generateGoal(pb);
  }

  private void generateGoal(Problem pb) {
    if (varSelector == null) varSelector = new MinDomain(pb);
    if (valIterator == null) valIterator = new IncreasingDomain();
    attachGoal(new AssignVar(varSelector, valIterator));
  }

  private void attachGoal(AbstractIntBranching branching) {
    branching.setSolver(solver);
    solver.mainGoal = branching;
  }

  public void addGoal(AbstractIntBranching branching) {
    branching.setSolver(solver);
    AbstractIntBranching br = solver.mainGoal;
    while(br.getNextBranching() != null) {
      br = (AbstractIntBranching) br.getNextBranching();
    }
    br.setNextBranching(branching);
  }

  /**
   * commands the solver to start
   */
  public void launch() {
    solver.run();
  }

  /**
   * returns the number of solutions encountered during the search
   * @return the number of solutions to the problem that were encountered during the search
   */
  public int getNbSolutions() {
    return solver.nbSolutions;
  }

  /**
   * Sets the time limit i.e. the maximal time before stopping the search algorithm
   */
  public void setTimeLimit(int timeLimit) {
    this.timeLimit = timeLimit;
  }

  /**
   * Sets the node limit i.e. the maximal number of nodes explored by the search algorithm
   */
  public void setNodeLimit(int nodeLimit) {
    this.nodeLimit = nodeLimit;
  }

  /**
   * @return true if only the first solution must be found
   */
  public boolean getFirstSolution() {
    return firstSolution;
  }

  /**
   * Sets wether only the first solution must be found
   */
  public void setFirstSolution(boolean firstSolution) {
    this.firstSolution = firstSolution;
  }

  /**
   * Sets the variable selector the search solver should use.
   */
  public void setVarSelector(IVarSelector varSelector) {
    this.varSelector = varSelector;
  }

  /**
   * Sets the value iterator the search should use
   */
  public void setValIterator(IValIterator valIterator) {
    this.valIterator = valIterator;
  }
}
