// *********************************************
// *                   J-CHOCO                 *
// *   Copyright (c) F. Laburthe, 1999-2003    *
// *********************************************
// * Event-base contraint programming Engine   *
// *********************************************

// CVS Information
// File:               $RCSfile: IntLinComb.java,v $
// Version:            $Revision: 1.3 $
// Last Modification:  $Date: 2004/06/30 06:45:22 $
// Last Contributor:   $Author: hadrien $

package choco.integer.constraints;

import choco.ContradictionException;
import choco.Problem;
import choco.AbstractConstraint;
import choco.Constraint;
import choco.integer.var.IntDomainVar;
import choco.integer.IntVar;
import choco.integer.IntExp;
import choco.util.Arithm;

// import choco.search.AbstractLargeIntConstraint;

/**
 * Implements a constraint Sigma (ai Xi) <=/>=/= C, with Xi variables, ai and C search constants.
 */
public class IntLinComb extends AbstractLargeIntConstraint {

  /**
   * Constant, to be assigned to <code>op</code>, representing linear equalities
   */
  public static final int EQ = 0;

  /**
   * Constant, to be assigned to <code>op</code>, representing linear inequalities
   */
  public static final int GEQ = 1;

  /**
   * Constant, to be assigned to <code>op</code>, representing linear disequalities
   */
  public static final int NEQ = 2;

  /**
   * Field representing the type of linear constraint (equality, inequality, disequality)
   */
  protected final int op;

  /**
   * Field representing the type of linear constraint (equality, inequality, disequality)
   */
  protected final int[] coeffs;

  /**
   * Field representing the number of variables with positive coeffficients in the linear combination
   * (all positive coefficients are given before negative ones)
   */
  protected final int nbPosVars;

  /**
   * The search constant of the constraint
   */
  protected final int cste;

  /**
   * rewriting utility: remove all null coefficients
   * @param lcoeffs
   * @return
   */
  // TODO: could be improved to remove duplicates (variables that would appear twice in the linear combination)
  private static int countNonNull(int[] lcoeffs) {
    // first, count the actual number of non-null coefficients
    int nbNonNull = 0;
    for (int i = 0; i < lcoeffs.length; i++) {
      if (lcoeffs[i] != 0)
        nbNonNull++;
    }
    return nbNonNull;
  }

  /**
   * Constructs the constraint with the specified variables and constant.
   *
   */
  public IntLinComb(choco.integer.IntVar[] lvars, int[] lcoeffs, int c, int linOperator) {
    // create the appropriate data structure
    super(countNonNull(lcoeffs));

    int nbVars = getNbVars();
    int i = 0;
    int j = 0;
    coeffs = new int[nbVars];
    // fill it up with the coefficients and variables in the right order
    for (i = 0; i < nbVars; i++) {
      if (lcoeffs[i] > 0) {
        vars[j] = lvars[i];
        coeffs[j] = lcoeffs[i];
        j++;
      }
    }
    nbPosVars = j;

    for (i = 0; i < nbVars; i++) {
      if (lcoeffs[i] < 0) {
        vars[j] = lvars[i];
        coeffs[j] = lcoeffs[i];
        j++;
      }
    }
    op = linOperator;
    cste = c;
  }

  public void propagate() throws ContradictionException {
    filter(true, 2);
  }

  public void awakeOnVar(int idx) throws ContradictionException {
  }

  public void awakeOnInf(int idx) throws ContradictionException {
    if (idx < nbPosVars)
      filter(true, 1);
    else
      filter(false, 1);
  }

  public void awakeOnSup(int idx) throws ContradictionException {
    if (idx < nbPosVars)
      filter(false, 1);
    else
      filter(true, 1);
  }

  public void awakeOnInst(int idx) throws ContradictionException {
    propagate();
  }

  public void awakeOnRem(int idx, int x) throws ContradictionException {
    ;
  }

  public Boolean isEntailed() {
    if (op == EQ) {
      int a = computeLowerBound();
      int b = computeUpperBound();
      if (b < 0 || a > 0)
        return Boolean.FALSE;
      else if (a == 0 && b == 0)
        return Boolean.TRUE;
      else
        return null;
    } else if (op == GEQ) {
      if (computeUpperBound() < 0)
        return Boolean.FALSE;
      else if (computeLowerBound() >= 0)
        return Boolean.TRUE;
      else
        return null;
    } else {
      assert(op == NEQ);
      int a = computeLowerBound();
      if (a > 0)
        return Boolean.TRUE;
      else {
        int b = computeUpperBound();
        if (b < 0)
          return Boolean.TRUE;
        else if (b == 0 && a == 0)
          return Boolean.FALSE;
        else
          return null;
      }
    }
  }


  public boolean isSatisfied() {
    int s = cste;
    int nbVars = getNbVars();
    int i;
    for (i = 0; i < nbVars; i++) {
      s += (vars[i].getValue() * coeffs[i]);
    }
    if (op == EQ)
      return (s == 0);
    else if (op == GEQ)
      return (s >= 0);
    else {
      assert(op == NEQ);
      return (s != 0);
    }
  }

  /**
   * computes an upper bound estimate of a linear combination of variables
   */
  protected int computeUpperBound() {
    int s = cste;
    int nbVars = getNbVars();
    int i;
    for (i = 0; i < nbPosVars; i++)
      s += (vars[i].getSup() * coeffs[i]);
    for (i = nbPosVars; i < nbVars; i++)
      s += (vars[i].getInf() * coeffs[i]);
    logger.finest("ub for " + toString() + ": " + s);
    return s;
  }

  /**
   * computes a lower bound estimate of a linear combination of variables
   */
  protected int computeLowerBound() {
    int s = cste;
    int nbVars = getNbVars();
    int i;
    for (i = 0; i < nbPosVars; i++)
      s += (vars[i].getInf() * coeffs[i]);
    for (i = nbPosVars; i < nbVars; i++)
      s += (vars[i].getSup() * coeffs[i]);
    logger.finest("lb for " + toString() + ": " + s);
    return s;
  }

// Note: additional propagation pass are sometimes useful:
// For instance : 3*X[0.3] + Y[1.10] = 10
//                Y >= 2 causes X < 3 -> updateSup(X,2)
//                and this very var (the new sup of X) causes (Y >= 4).
//                this induced var (Y>=4) could not be infered at first (with only Y>=2)
//
  /**
   *  a strategy for chaotic iteration with two rules (LB and UB propagation)
   *  the fix point is reached individually for each rule in one function call
   *  but this call may break the stability condition for the other rule (in which case
   *  the second rule infers new information from the fresh inferences made by the first rule).
   *  The algorithm oscilates between both rules until a global fix point is reached.
   *
   * @param startWithLB whether LB must be the first rule applied
   * @param minNbRules minimum number of rules required to reach fiux point.
   */
  protected void filter(boolean startWithLB, int minNbRules) throws ContradictionException {
    boolean lastRuleEffective = true;     // whether the last rule indeed perform some reductions
    int nbr = 0;                          // number of rules applied
    boolean nextRuleIsLB = startWithLB;   // whether the next rule that should be filtered is LB (or UB)
    while (lastRuleEffective || nbr < minNbRules) {
      if (nextRuleIsLB) {
        logger.finer("-- LB propagation for " + toString());
        lastRuleEffective = filterOnImprovedLowerBound();
      } else {
        logger.finer("-- UB propagation for " + toString());
        lastRuleEffective = filterOnImprovedUpperBound();
      }
      nextRuleIsLB = !nextRuleIsLB;
      nbr++;
    }
  }

  protected boolean filterOnImprovedLowerBound() throws ContradictionException {
    if (op == EQ)    // the constraint check is needed only for equality constraints (otherwise passive constraint)
      return propagateNewLowerBound(computeLowerBound());
    else if (op == GEQ)
      return false;   // nothing to propagate => nothing was inferred, so return false
    else {
      assert(op == NEQ);
      int mylb = computeLowerBound();
      if (mylb == 0) { // propagate the constraint sigma(ai) + c >= 1
        logger.finer("propagate > 0 for " + toString());
        return propagateNewUpperBound(computeUpperBound() - 1); // TODO I have doubts.....
      } else
        return false;
    }
  }

  protected boolean filterOnImprovedUpperBound() throws ContradictionException {
    int myub = computeUpperBound();
    if (op == EQ)
      return propagateNewUpperBound(myub);
    else if (op == GEQ)
      return propagateNewUpperBound(myub);
    else {
      assert(op == NEQ);
      if (myub == 0) { // propagate the constraint sigma(ai) + c <= -1
        logger.finer("propagate < 0 for " + toString());
        return propagateNewLowerBound(computeLowerBound() + 1); // TODO I have doubts.....
      } else
        return false;
    }
  }

  /**
   * propagates the constraint sigma(ai Xi) + c <= 0 where mylb = sigma(ai inf(Xi)) + c
   *  note: this does not reach saturation (fix point),
   *    but returns a boolean indicating whether it infered new information or not.
   */
  protected boolean propagateNewLowerBound(int mylb) throws ContradictionException {
    boolean anyChange = false;
    int nbVars = getNbVars();
    if (mylb > 0) {
      logger.finer("lb = " + mylb + " > 0 => fail");
      this.fail();
    }
    int i;
    for (i = 0; i < nbPosVars; i++) {
      int newSupi = Arithm.divFloor(-(mylb), coeffs[i]) + vars[i].getInf();
      if (vars[i].updateSup(newSupi, cIndices[i])) {
        logger.finer("SUP(" + vars[i].toString() + ") <= " + -(mylb) + "/" + coeffs[i] + " + " + vars[i].getInf() + " = " + newSupi);
        anyChange = true;
      }
    }
    for (i = nbPosVars; i < nbVars; i++) {
      int newInfi = Arithm.divCeil(mylb, -(coeffs[i])) + vars[i].getSup();
      if (vars[i].updateInf(newInfi, cIndices[i])) {
        logger.finer("INF(" + vars[i].toString() + ") >= " + mylb + "/" + -(coeffs[i]) + " + " + vars[i].getSup() + " = " + newInfi);
        anyChange = true;
      }
    }
    return anyChange;
  }

  /**
   * propagates the constraint sigma(ai Xi) + c <= 0 where myub = sigma(ai sup(Xi)) + c
   *  note: this does not reach saturation (fix point),
   *    but returns a boolean indicating whether it infered new information or not.
   */
  protected boolean propagateNewUpperBound(int myub) throws ContradictionException {
    boolean anyChange = false;
    int nbVars = getNbVars();
    if (myub < 0) {
      logger.finer("ub = " + myub + " < 0 => fail");
      this.fail();
    }
    int i;
    for (i = 0; i < nbPosVars; i++) {
      int newInfi = Arithm.divCeil(-(myub), coeffs[i]) + vars[i].getSup();
      if (vars[i].updateInf(newInfi, cIndices[i])) {
        logger.finer("INF(" + vars[i].toString() + ") >= " + -(myub) + "/" + coeffs[i] + " + " + vars[i].getSup() + " = " + newInfi);
        anyChange = true;
      }
    }
    for (i = nbPosVars; i < nbVars; i++) {
      int newSupi = Arithm.divFloor(myub, -(coeffs[i])) + vars[i].getInf();
      if (vars[i].updateSup(newSupi, cIndices[i])) {
        logger.finer("SUP(" + vars[i].toString() + ") <= " + myub + "/" + -(coeffs[i]) + " + " + vars[i].getInf() + " = " + newSupi);
        anyChange = true;
      }
    }
    return anyChange;
  }

  /**
   * tests if the constraint is consistent with respect to the current state of domains
   * @return true iff the constraint is bound consistent (weaker than arc consistent)
   */
  public boolean isConsistent() {
    int ub = computeUpperBound();
    int lb = computeLowerBound();
    int nbVars = getNbVars();

    if (op == EQ) {
      return (hasConsistentLowerBound() && hasConsistentUpperBound());
    } else if (op == GEQ) {
      return hasConsistentUpperBound();
    }
    return true;
  }

  /**
   * tests if the constraint is consistent with respect to the current state of domains
   * @return true iff the constraint is bound consistent (weaker than arc consistent)
   */
  protected boolean hasConsistentLowerBound() {
    int lb = computeLowerBound();
    int nbVars = getNbVars();

    if (lb > 0) {
      return false;
    } else {
      for (int i = 0; i < nbPosVars; i++) {
        int newSupi = Arithm.divFloor(-(lb), coeffs[i]) + vars[i].getInf();
        if (vars[i].getSup() < newSupi) {
          return false;
        }
      }
      for (int i = nbPosVars; i < nbVars; i++) {
        int newInfi = Arithm.divCeil(lb, -(coeffs[i])) + vars[i].getSup();
        if (vars[i].getInf() > newInfi) {
          return false;
        }
      }
      return true;
    }
  }

  protected boolean hasConsistentUpperBound() {
    int ub = computeUpperBound();
    int nbVars = getNbVars();

    if (ub < 0) {
      return false;
    } else {
      for (int i = 0; i < nbPosVars; i++) {
        int newInfi = Arithm.divCeil(-(ub), coeffs[i]) + vars[i].getSup();
        if (vars[i].getInf() > newInfi) {
          return false;
        }
      }
      for (int i = nbPosVars; i < nbVars; i++) {
        int newSupi = Arithm.divFloor(ub, -(coeffs[i])) + vars[i].getInf();
        if (vars[i].getSup() < newSupi) {
          return false;
        }
      }
      return true;
    }
  }

  public AbstractConstraint opposite() {
    Problem pb = getProblem();
    IntExp term = pb.scalar(coeffs, vars);
    if (op == EQ) {
      return (AbstractConstraint) pb.neq(term, cste);
    } else if (op == NEQ) {
      return (AbstractConstraint) pb.eq(term, cste);
    } else if (op == GEQ) {
      return (AbstractConstraint) pb.lt(term, cste);
    } else {
      return null;
    }
  }

  // TODO: we need to sort the variables and coefficients
  public boolean isEquivalentTo(Constraint compareTo) {
    if (compareTo instanceof IntLinComb) {
      IntLinComb c = (IntLinComb) compareTo;
      int n = this.getNbVars();
      if (c.getNbVars() == n) {
        boolean allEqual = (this.op == c.op);
        for (int varIdx = 0; varIdx < n; varIdx++) {
          if ( (this.getIntVar(varIdx) != c.getIntVar(varIdx)) ||
               (this.coeffs[varIdx] != c.coeffs[varIdx]) ) {
            allEqual = false;
            break;
          }
        }
        return allEqual;
      } else {
        return false;
      }
    } else {
      return false;
    }
  }

  public String toString() {
    String linComb = "";
    for (int i = 0; i < coeffs.length - 1; i++) {
     linComb += coeffs[i] + "*" + vars[i] + " + " ;
    }
    linComb += coeffs[coeffs.length - 1] + "*" + vars[coeffs.length - 1];
    if (op == 0) linComb += " = ";
    else if (op == 1) linComb += " >= ";
    else if (op == 2) linComb += " != ";
    linComb += - cste;
    return linComb;
  }
}
