// *********************************************
// *                   J-CHOCO                 *
// *   Copyright (c) F. Laburthe, 1999-2003    *
// *********************************************
// * Event-base contraint programming Engine   *
// *********************************************

// CVS Information
// File:               $RCSfile: IntDomainVar.java,v $
// Version:            $Revision: 1.1.1.1 $
// Last Modification:  $Date: 2004/05/11 13:51:47 $
// Last Contributor:   $Author: laburthe $

package choco.integer.var;

import choco.AbstractVar;
import choco.Constraint;
import choco.ContradictionException;
import choco.Problem;
import choco.integer.var.*;
import choco.integer.IntVar;
import choco.mem.Environment;
import choco.mem.StoredInt;
import choco.prop.VarEvent;
import choco.util.IntIterator;

import java.util.logging.Logger;

/**
 * Implements search valued domain variables.
 */
public class IntDomainVar extends AbstractVar implements IntVar {

  /**
   * The backtrackable value of the variable, if instantiated.
   */

  public final StoredInt value; // voir si il faut le storer...


  /**
   * The backtrackable domain of the variable.
   */

  protected AbstractIntDomain domain;

  /**
   * Reference to an object for logging trace statements related to IntDomainVar (using the java.util.logging package)
   */

  protected static Logger logger = Logger.getLogger("choco.search.IntDomainVar");

  /**
   * Constructs a new variable for the specified problem and with the
   * specified name and bounds.
   * @param pb The problem of the variable.
   * @param name Its name.
   * @param domainType the type of encoding for the domain (BOUNDS, LIST, ...)
   * @param a Its minimal value.
   * @param b Its maximal value.
   */

  public IntDomainVar(Problem pb, String name, int domainType, int a, int b) {
    Environment env = pb.getEnvironment();

    this.problem = pb;
    this.name = name;
    if (domainType == IntVar.LIST) {
      domain = new BitSetIntDomain(this, a, b);
    } else {
      domain = new IntervalIntDomain(this, a, b);
    }
    this.value = (StoredInt) env.makeInt();
    if (a == b) this.value.set(a);
    this.event = new IntVarEvent(this);
  }



  // ============================================
  // Methods of the interface
  // ============================================

  /**
   * Checks if the variable is instantiated to a specific value.
   */

  public boolean isInstantiatedTo(int x) {
    return (getValue() == x);
  }


  /**
   * Checks if the variables is instantiated to any value.
   */

  public boolean isInstantiated() {
    return (value.isKnown());
  }


  /**
   * Checks if a value is still in the domain.
   */

  public boolean canBeInstantiatedTo(int x) {
    return (getInf() <= x &&
        x <= getSup() &&
        (domain == null || domain.contains(x)));
  }


  /**
   * Sets the minimum value.
   */

  public void setMin(int x) throws ContradictionException {
    updateInf(x, VarEvent.NOCAUSE);
  }


  /**
   * Sets the maximal value.
   */

  public void setMax(int x) throws ContradictionException {
    updateSup(x, VarEvent.NOCAUSE);
  }


  /**
   * Instantiates the variable.
   * @throws choco.ContradictionException
   */

  public void setVal(int x) throws ContradictionException {
    instantiate(x, VarEvent.NOCAUSE);
  }


  /**
   * Removes a value.
   * @throws choco.ContradictionException
   */

  public void remVal(int x) throws ContradictionException {
    removeVal(x, VarEvent.NOCAUSE);
  }

  public void wipeOut() throws ContradictionException {
    throw new ContradictionException(this);
  }

  public boolean hasEnumeratedDomain() {
    return domain.isEnumerated();
  }

  public IntDomain getDomain() {
    return domain;
  }

  /**
   * Gets the domain size.
   */

  public int getDomainSize() {
        return domain.getSize();
    }

  /**
   * Checks if it can be equals to another variable.
   */

  public boolean canBeEqualTo(IntVar x) {
    if (x.getInf() <= this.getSup()) {
      if (this.getInf() <= x.getSup()) {
        if (!this.hasEnumeratedDomain() && !x.hasEnumeratedDomain())
          return true;
        else {
          for (IntIterator it = this.getDomain().getIterator(); it.hasNext();) {
            int v = it.next();
            if (x.canBeInstantiatedTo(v))
              return true;
          }
          return false;
        }
      } else
        return false;
    } else
      return false;
  }


  /**
   * Checks if the variables can be instantiated to at least one value
   * in the array.
   * @param sortedValList The value array.
   * @param nVals The number of interesting value in this array.
   */

  public boolean canBeInstantiatedIn(int[] sortedValList, int nVals) {
    if (getInf() <= sortedValList[nVals - 1]) {
      if (getSup() >= sortedValList[0]) {
        if (domain == null)
          return true;
        else {
          for (int i = 0; i < nVals; i++) {
            if (canBeInstantiatedTo(sortedValList[i]))
              return true;
          }
          return false;
        }
      } else
        return false;
    } else
      return false;
  }


  /**
   * Returns a randomly choosed value in the domain.
   *
   * Not implemented yet.
   */

  public int getRandomDomainValue() {
    if (domain == null)
      return getInf();
// TODO     return inf.get() + random(sup.get() - inf.get() + 1);
    else
      return domain.getRandomValue();
  }


  /**
   * Gets the next value in the domain.
   */

  public int getNextDomainValue(int currentv) {
    if (currentv < getInf())
      return getInf();
    else if (domain == null)
      return currentv + 1;
    else
      return domain.getNextValue(currentv);
  }


  /**
   * Gets the previous value in the domain.
   */

  public int getPrevDomainValue(int currentv) {
    if (currentv > getSup())
      return getSup();
    else if (domain == null)
      return currentv - 1;
    else
      return domain.getPrevValue(currentv);
  }


  /**
   * Internal var: update on the variable lower bound caused by its i-th
   * constraint.
   * Returns a boolean indicating whether the call indeed added new information
   * @param x    The new lower bound.
   * @param idx  The index of the constraint (among all constraints linked to
   * the variable) responsible for the update.
   */

  public boolean updateInf(int x, int idx) throws ContradictionException {
    logger.finest("INF(" + this.toString() + "): " + this.getInf() + " -> " + x);
    return domain.updateInf(x, idx);
  }

  /**
   * Internal var: update on the variable upper bound caused by its i-th
   * constraint.
   * Returns a boolean indicating whether the call indeed added new information.
   * @param x    The new upper bound
   * @param idx  The index of the constraint (among all constraints linked to
   * the variable) responsible for the update
   */

  public boolean updateSup(int x, int idx) throws ContradictionException {
    logger.finest("SUP(" + this.toString() + "): " + this.getSup() + " -> " + x);
    return domain.updateSup(x, idx);
  }


  /**
   * Internal var: update (value removal) on the domain of a variable caused by
   * its i-th constraint.
   * <i>Note:</i> Whenever the hole results in a stronger var (such as a bound update or
   *    an instantiation, then we forget about the index of the var generating constraint.
   *    Indeed the propagated var is stronger than the initial one that
   *    was generated; thus the generating constraint should be informed
   *    about such a new var.
   * Returns a boolean indicating whether the call indeed added new information.
   * @param x    The removed value
   * @param idx  The index of the constraint (among all constraints linked to the variable) responsible for the update
   */

  public boolean removeVal(int x, int idx) throws ContradictionException {
    return domain.removeVal(x, idx);
  }

  /**
   * Internal var: remove an interval (a sequence of consecutive values) from
   * the domain of a variable caused by its i-th constraint.
   * Returns a boolean indicating whether the call indeed added new information.
   * @param a 	the first removed value
   * @param b 	the last removed value
   * @param idx  the index of the constraint (among all constraints linked to the variable)
   * responsible for the update
   */

  public boolean removeInterval(int a, int b, int idx) throws ContradictionException {
    return domain.removeInterval(a, b, idx);
  }

  /**
   * Internal var: instantiation of the variable caused by its i-th constraint
   * Returns a boolean indicating whether the call indeed added new information.
   * @param x 	the new upper bound
   * @param idx  the index of the constraint (among all constraints linked to the
   * variable) responsible for the update
   */

  public boolean instantiate(int x, int idx) throws ContradictionException {
    return domain.instantiate(x, idx);
  }


  public void fail() throws ContradictionException {
    problem.getChocEngine().raiseContradiction(this);
  }


  /**
   * Gets the minimal value of the variable.
   */

  public int getInf() {
    return domain.getInf();
  }


  /**
   * Gets the maximal value of the variable.
   */

  public int getSup() {
    return domain.getSup();
  }


  /**
   * Gets the value of the variable if instantiated.
   */

  public int getValue() {
    return value.get();
  }

  /**
   * pretty printing
   * @return a String representation of the variable
   */
  public String pretty() {
    return (this.toString() + this.domain.pretty());
  }
}
