package choco.mem;

import java.util.logging.Logger;

/**
 * A backtrackable float variable trail storing past values of all the float variables.
 */
public class StoredFloatTrail implements ITrailStorage {
    /**
   * Reference to an object for logging trace statements related memory & backtrack (using the java.util.logging package)
   */

  private static Logger logger = Logger.getLogger("choco.mem");

  /**
   * Reference towards the overall environment
   * (responsible for all memory management).
   */

  private Environment environment;


  /**
   * Stack of backtrackable search variables.
   */

  private StoredFloat[] variableStack;


  /**
   * Stack of values (former values that need be restored upon backtracking).
   */

  private double[] valueStack;


  /**
   * Stack of timestamps indicating the world where the former value
   * had been written.
   */

  private int[] stampStack;


  /**
   * Points the level of the last entry.
   */

  private int currentLevel;


  /**
   * A stack of pointers (for each start of a world).
   */

  private int[] worldStartLevels;

  /**
   * capacity of the trailing stack (in terms of number of updates that can be stored)
   */
  private int maxUpdates = 0;

  /**
   * capacity of the trailing stack (in terms of number of worlds that can be handled)
  */
  private int maxWorlds = 0;

  /** Constructs a trail with predefined size.
   * @param nUpdates maximal number of updates that will be stored
   * @param nWorlds  maximal number of worlds that will be stored
   */

  public StoredFloatTrail(Environment env, int nUpdates, int nWorlds) {
    environment = env;
    currentLevel = 0;
    maxUpdates = nUpdates;
    maxWorlds = nWorlds;
    variableStack = new StoredFloat[nUpdates];
    valueStack = new double[nUpdates];
    stampStack = new int[nUpdates];
    worldStartLevels = new int[nWorlds];
  }


  /**
   * Moving up to the next world.
   */

  public void worldPush() {
    worldStartLevels[environment.currentWorld + 1] = currentLevel;
  }


  /**
   * Moving down to the previous world.
   */

  public void worldPop() {
    while (currentLevel > worldStartLevels[environment.currentWorld]) {
      currentLevel--;
      StoredFloat v = variableStack[currentLevel];
      v._set(valueStack[currentLevel], stampStack[currentLevel]);
    }
  }


  /**
   * Returns the current size of the stack.
   */

  public int getSize() {
    return currentLevel;
  }


  /**
   * Comits a world: merging it with the previous one.
   */

  public void worldCommit() {
    // TODO
  }


  /**
   * Reacts when a StoredInt is modified: push the former value & timestamp
   * on the stacks.
   */

  public void savePreviousState(StoredFloat v, double oldValue, int oldStamp) {
    valueStack[currentLevel] = oldValue;
    variableStack[currentLevel] = v;
    stampStack[currentLevel] = oldStamp;
    currentLevel++;
    if (currentLevel == maxUpdates)
      resizeUpdateCapacity();
  }

 private void resizeUpdateCapacity() {
   int newCapacity =  ((maxUpdates * 3) / 2);
   // first, copy the stack of variables
   StoredFloat[] tmp1 = new StoredFloat[newCapacity];
   System.arraycopy(variableStack, 0, tmp1, 0, variableStack.length);
   variableStack = tmp1;
   // then, copy the stack of former values
   double[] tmp2 = new double[newCapacity];
   System.arraycopy(valueStack, 0, tmp2, 0, valueStack.length);
   valueStack = tmp2;
   // then, copy the stack of world stamps
   int[] tmp3 = new int[newCapacity];
   System.arraycopy(stampStack, 0, tmp3, 0, stampStack.length);
   stampStack = tmp3;
   // last update the capacity
   maxUpdates = newCapacity;
 }

  public void resizeWorldCapacity(int newWorldCapacity) {
    int[] tmp = new int[newWorldCapacity];
    System.arraycopy(worldStartLevels, 0, tmp, 0, worldStartLevels.length);
    worldStartLevels = tmp;
    maxWorlds = newWorldCapacity;
 }
}
