//     ___  ___         PaLM library for Choco system
//    /__ \/ __\        (c) 2001 - 2004 -- Narendra Jussien
//   ____\ |/_____
//  /_____\/_____ \     Explanation based constraint programming tool
// |/   (_)(_)   \|
//       \ /            Version 0.1
//       \ /            January 2004
//       \ /
// ______\_/_______     Contibutors: Francois Laburthe, Hadrien Cambazard, Guillaume Rochart...

package choco.palm;

import choco.Solver;
import choco.Problem;
import choco.Solution;
import choco.integer.IntVar;
import choco.integer.var.IntDomainVar;
import choco.search.BranchAndBound;
import choco.search.NodeLimit;
import choco.palm.search.*;
import choco.palm.search.pathrepair.PathRepairLearn;
import choco.palm.search.pathrepair.PathRepairAssignVar;

import java.util.ArrayList;
import java.util.List;

/**
 * This class extends Solver as a factory for PaLM search solvers.
 */
public class PalmSolver extends Solver {
  /**
   * The extender class the generated solver should use to choose decision constraints during search.
   */
  protected PalmExtend extend = null;

  /**
   * The state class the generated solver should use to maintain the state of the search.
   */
  protected PalmState state = null;

  /**
   * The learning class the generated solver should use to learn forbidden or authorized nodes during search.
   */
  protected PalmLearn learn = null;

  /**
   * The repair class the generated solver should use to choose which decision constraints to remove for repairing.
   */
  protected PalmRepair repair = null;

  /**
   * Lists of the branchings that should be used by the generated solver.
   */
  protected List branchings = new ArrayList();

  /**
   * <i>Decision Repair</i>: number of explanations the decision repair algorithm should store.
   */
  protected int prSize = -1;
  /**
   * <i>Decision Repair</i>: number of moves without improvment the decision repair algorithm should try before
   * stopping.
   */
  protected int prMoves = -1;

  /**
   * Creates a new solver factory for the specified problem
   * @param pb
   */
  public PalmSolver(Problem pb) {
    super(pb);
  }

  /**
   * The factory method: builds the solver needed to solve the problem (an optimization solver if a variable
   * should be minimized or maximized, a classical solver (mac-dbt) for constraint problems, or a path-repair
   * or decision-repair solver if explanations should be kept.
   * @param pb The problem the solver should solve.
   */
  public void generateSearchSolver(Problem pb) {
    problem = pb; // TODO : a priori inutile !
    if (null == objective && prSize < 0) {                   // MAC-DBT
      solver = new PalmGlobalSearchSolver(pb);
    } else if (null != objective && prSize < 0) {            // MAC-DBT + Dynamic Cuts
      solver = new PalmBranchAndBound(pb, objective, doMaximize);
    } else if (null == objective && prSize >= 0) {           // Decision repair
      solver = new PalmGlobalSearchSolver(pb);
      if (extend == null) extend = new PalmUnsureExtend();
      if (repair == null) repair = new PalmUnsureRepair();
      if (learn == null) learn = new PathRepairLearn(prSize, prMoves);
      if (branchings.size() == 0) branchings.add(new PathRepairAssignVar());
    }

    // Classical solver tools
    if (state == null) state = new PalmState((((PalmProblem)this.problem).makeExplanation()));
    if (repair == null) repair = new PalmRepair();
    if (learn == null) learn = new PalmLearn();
    if (extend == null) extend = new PalmExtend();

    // Classical limits
    solver.limits.add(new PalmTimeLimit(solver, timeLimit));
    solver.limits.add(new NodeLimit(solver, nodeLimit));
    // TODO : limits.add(new relaxLimit());

    // Solver should stop at first solution ? TODO : see if useful !
    solver.stopAtFirstSol = firstSolution;

    // Attach solver tools
    ((PalmGlobalSearchSolver) solver).attachPalmState(state);
    ((PalmGlobalSearchSolver) solver).attachPalmExtend(extend);
    ((PalmGlobalSearchSolver) solver).attachPalmLearn(learn);
    ((PalmGlobalSearchSolver) solver).attachPalmRepair(repair);

    // Attach branchings (with a default one if needed
    if (branchings.size() == 0) {
      branchings.add(new PalmAssignVar());
    }
    ((PalmGlobalSearchSolver) solver).attachPalmBranchings(branchings);
  }

  /**
   * Attaches a list of goals -- or branching in the case of Palm -- that should be attched
   * to the generated solver.
   * @param lst list of branchings (PalmAbstractBranching)
   */
  public void attachGoals(List lst) { // TODO : dans Choco ?
    branchings = lst;
  }

  /**
   * Adds a new branching to the branching list to attach to the generated solver.
   * @param br
   */
  public void addGoal(PalmAbstractBranching br) { // TODO : utiliser meme methode que CHoco ?
    branchings.add(br);
  }

  /**
   * Return the <code>i</code>th solutions.
   * @param i
   * @return
   */
  public Solution getSolution(int i) {  // TODO : dans Choco !!
    return (Solution) solver.solutions.get(i);
  }

  /**
   * Sets the objectove to minimmize or maximize.
   * @param var The optimization variable
   */
  public void setObjective(IntVar var) { // TODO : dans Choco !!
    this.objective = (IntDomainVar) var;
  }

  /**
   * States that the solve should maximize the objective variable.
   */
  public void setMaximize() { // TODO : dans Choco !!
    this.doMaximize = true;
  }

  /**
   * States that the solve should minimize the objective variable.
   */
  public void setMinimize() { // TODO : dans Choco !!
    this.doMaximize = false;
  }

  /**
   * Sets a custom extender to attach to the generated solver.
   */
  public void setPalmExtender(PalmExtend ext) {
    extend = ext;
  }

  /**
   * Sets a custom learner to attach to the generated solver.
   */
  public void setPalmLearner(PalmLearn ext) {
    learn = ext;
  }

  /**
   * Sets a custom state to attach to the generated solver.
   */
  public void setPalmState(PalmState ext) {
    state = ext;
  }

  /**
   * Sets a custom repairer to attach to the generated solver.
   */
  public void setPalmRepairer(PalmRepair ext) {
    repair = ext;
  }

  /**
   * <i>Decision Repair</i> Sets the maximal number of explanations stored and the maximal moves without
   * improvement of the decision repair algorithm.
   * @param size
   * @param moves
   */
  public void setPathRepairValues(int size, int moves) {
    this.prSize = size;
    this.prMoves = moves;
  }
}
