//     ___  ___         PaLM library for Choco system
//    /__ \/ __\        (c) 2001 - 2004 -- Narendra Jussien
//   ____\ |/_____
//  /_____\/_____ \     Explanation based constraint programming tool
// |/   (_)(_)   \|
//       \ /            Version 0.1
//       \ /            January 2004
//       \ /
// ______\_/_______     Contibutors: Franois Laburthe, Hadrien Cambazard, Guillaume Rochart...

package choco.palm.search;

import choco.palm.PalmProblem;
import choco.palm.integer.constraints.PalmGreaterOrEqualXC;
import choco.palm.integer.constraints.PalmLessOrEqualXC;

import java.util.LinkedList;
import java.util.logging.Logger;
import java.util.logging.Level;

import choco.integer.IntVar;
import choco.ContradictionException;
import choco.Constraint;
import choco.Problem;

/**
 * An optimizing solver.
 */

public class PalmBranchAndBound extends PalmGlobalSearchSolver {

  /**
   * States if the solver is maximizing (or minimizing).
   */

  protected boolean maximizing = false;


  /**
   * The variable that should be maximized (or minimized).
   */

  protected IntVar objective;


  /**
   * Some bounds for search use.
   */

  protected int lowerBound = Integer.MIN_VALUE, upperBound = Integer.MAX_VALUE;


  /**
   * Optimum value found during the search.
   */

  protected int optimum;


  /**
   * Dynamic cut constraints.
   */

  protected LinkedList dynamicCuts;


  /**
   * Creates the solver for the specified problem.
   */

  public PalmBranchAndBound(Problem pb, IntVar obj, boolean max) {
    super(pb);
    objective = obj;
    maximizing = max;
    lowerBound = obj.getInf();
    upperBound = obj.getSup();
    dynamicCuts = new LinkedList();
  }

  public void run() {
    try {
      problem.post(problem.geq(objective,lowerBound));
      problem.post(problem.leq(objective,upperBound));
      if (Logger.getLogger("choco.palm").isLoggable(Level.INFO))
        Logger.getLogger("choco.palm").info("Initial Propagation");
      problem.propagate();
    } catch (ContradictionException e) {
      if (Logger.getLogger("choco.palm").isLoggable(Level.INFO))
        Logger.getLogger("choco.palm").info("Optimality proven");
      this.finished = true;
      this.problem.feasible  = Boolean.FALSE;
    }
    this.problem.feasible  = Boolean.TRUE;
    while (this.problem.feasible.booleanValue()) {
      if (Logger.getLogger("choco.palm").isLoggable(Level.INFO))
        Logger.getLogger("choco.palm").info("Searching for one solution");
      runonce();
      recordSolution();
      if (this.problem.feasible.booleanValue()) postDynamicCut();
    }
  }

  /**
   * solving the feasibility problem
   * same code as in run(PalmGlobalSearchSolver) in order to be able
   * to store consecutive solutions ...
   */
  public void runonce() {
    //long time = System.currentTimeMillis();
    try {
      while (!this.finished) {
        try {
          if (Logger.getLogger("choco.palm").isLoggable(Level.FINE))
            Logger.getLogger("choco.palm").fine("New extension launched.");
          this.extend();
          this.problem.propagate();
        } catch (PalmContradiction e) {
          this.repair();
        }
      }
      this.problem.feasible = Boolean.TRUE;
    } catch (ContradictionException e) {
      this.finished = true;
      this.problem.feasible = Boolean.FALSE;
    }
    //this.setRuntimeStatistic(PalmProblem.CPU, (int) (System.currentTimeMillis() - time));
  }

  public void postDynamicCut() {
    int bv = getObjectiveValue();
    Constraint cut;
    try {
       reset();
       if (maximizing) lowerBound = Math.max(lowerBound,bv) + 1;
       else upperBound = Math.min(upperBound,bv) - 1;
       if (maximizing) cut = new PalmGreaterOrEqualXC(objective,bv + 1);
       else cut = new PalmLessOrEqualXC(objective,bv - 1);
       dynamicCuts.add(cut);
       problem.post(cut);
       try {
         problem.propagate();
       } catch (PalmContradiction e) {
         repair();
       }
    } catch (ContradictionException e) {
       this.finished = true;
       this.problem.feasible = Boolean.FALSE;
    }
  }

  public int getObjectiveValue() {
    if (maximizing) return objective.getSup();
    else return objective.getInf();
  }

}
