package choco.real.constraint;

import choco.AbstractConstraint;
import choco.Var;
import choco.AbstractVar;
import choco.ContradictionException;
import choco.palm.explain.Explanation;
import choco.palm.explain.PalmConstraintPlugin;
import choco.palm.PalmProblem;
import choco.palm.integer.PalmIntVar;
import choco.util.IntIterator;
import choco.bool.AbstractCompositeConstraint;
import choco.integer.IntVar;
import choco.real.RealVar;
import choco.real.exp.RealIntervalConstant;

/**
 * Let x be an integer variable with n values and v be a real variable. Given n constant values a1 to an,
 * this constraint ensures that:
 *
 * <code>x = i iff v = ai</code>
 *
 * a1... an sequence is supposed to be ordered (a1&lt;a2&lt;... an)
 */
public class MixedCstElt extends AbstractConstraint implements MixedConstraint {
  protected RealVar realVar;
  protected int realIdx;
  protected IntVar intVar;
  protected int intIdx;
  protected double[] values;

  public MixedCstElt(RealVar v0, IntVar v1, double[] values) {
    this.problem = v0.getProblem();
    this.realVar = v0;
    this.intVar = v1;
    this.values = values;
  }

  public void awake() throws ContradictionException {
    intVar.updateSup(values.length - 1, intIdx);
    intVar.updateInf(0, intIdx);
    this.propagate();
  }

  public void propagate() throws ContradictionException {
    updateIInf();
    updateISup();
    updateReal();
  }

  public void awakeOnInf(int idx) throws ContradictionException {
    if (idx == 0) {
      updateIInf();
      updateReal();
    } else {
      updateReal();
    }
  }

  public void awakeOnSup(int idx) throws ContradictionException {
    if (idx == 0) {
      updateISup();
      updateReal();
    } else {
      updateReal();
    }
  }

  public void awakeOnBounds(int idx) throws ContradictionException {
    if (idx == 0) {
      updateIInf();
      updateISup();
      updateReal();
    } else {
      updateReal();
    }
  }

  public void awakeOnInst(int varIdx) throws ContradictionException {
    awakeOnBounds(varIdx);
  }

  public void awakeOnRem(int varIdx, int val) throws ContradictionException {
  }

  public void awakeOnRemovals(int varIdx, IntIterator deltaDomain) throws ContradictionException {
  }

  public void updateIInf() throws ContradictionException {
    int inf = intVar.getInf();
    while (values[inf] < realVar.getInf()) {
      inf++;
    }
    if (inf > intVar.getSup()) problem.getPropagationEngine().raiseContradiction();
    intVar.updateInf(inf, intIdx);
  }

  public void updateISup() throws ContradictionException {
    int sup = intVar.getSup();
    while (values[sup] > realVar.getSup()) {
      sup--;
    }
    if (sup < intVar.getInf()) problem.getPropagationEngine().raiseContradiction();
    intVar.updateSup(sup, intIdx);
  }

  public void updateReal() throws ContradictionException {
    realVar.intersect(new RealIntervalConstant(values[intVar.getInf()], values[intVar.getSup()]));
  }

  public boolean isConsistent() {
    return values[intVar.getInf()] <= realVar.getSup() && realVar.getInf() <= values[intVar.getSup()];
  }

  public boolean isSatisfied() {
    return isConsistent();
  }

  // Variable management
  public RealVar getRealVar(int i) {
    if (i == 0) return realVar;
    return null;
  }

  public int getRealVarNb() {
    return 1;
  }

  public IntVar getIntVar(int i) {
    if (i == 0) return intVar;
    return null;
  }

  public int getIntVarNb() {
    return 1;
  }

  public int getNbVars() {
    return 2;
  }

  public Var getVar(int i) {
    if (i == 0) return realVar;
    if (i == 1) return intVar;
    return null;
  }

  public void setConstraintIndex(int i, int idx) {
    if (i == 0) realIdx = idx;
    else if (i == 1) intIdx = idx;
    else throw new Error("bug in setConstraintIndex i:" + i + " this: " + this);
  }

  public int getConstraintIdx(int idx) {
    if (idx == 0) return realIdx;
    if (idx == 1) return intIdx;
    return -1;
  }

  public int assignIndices(AbstractCompositeConstraint root, int i) {
    int j=i;
    for (int k=0; k<getNbVars(); k++) {
      j++;
      connectVar((AbstractVar) getVar(k), j);
      setConstraintIndex(k, getVar(k).getNbConstraints() - 1);
    }
    return j;
  }

  public boolean isCompletelyInstantiated() {
    return intVar.isInstantiated() && realVar.isInstantiated();
  }
}
