package choco.search;

import choco.AbstractSolver;
import choco.ContradictionException;
import choco.Problem;
import choco.branch.AbstractBranching;

import java.util.ArrayList;
import java.util.logging.Level;

// **************************************************
// *                   J-CHOCO                      *
// *   Copyright (C) F. Laburthe, 1999-2003         *
// **************************************************
// *  an open-source Constraint Programming Kernel  *
// *     for Research and Education                 *
// **************************************************

/* File choco.search.AbstractGlobalSearchSolver.java, last modified by flaburthe 21 oct. 2003 09:52:56 */

/**
 * An abstract class for controlling tree search in various ways
 */
public abstract class AbstractGlobalSearchSolver extends AbstractSolver {

  /**
   * constants for dirving the incremental search algorithm
   */
  public static final int INIT_SEARCH = -1;
  public static final int OPEN_NODE = 0;
  public static final int UP_BRANCH = 1;
  public static final int DOWN_BRANCH = 2;

  /** a data structure storing the stack of choice contexts (for incremental search explorations) */
  public ArrayList traceStack = null;

  /** index of the current trace in the stack  */
  public int currentTraceIndex = -1;

  /** search controller: a flag storing the next move in the search tree */
  public int nextMove = INIT_SEARCH;

  /**
   * indicates whether the control should stop after the first solution is found
   */
  public boolean stopAtFirstSol = true;

  /**
   *  indicates whether a limit was encountered in the alst incremental search
   */
  protected boolean encounteredLimit = false;

  /**
   * count of the solutions found during search
   */
  public int nbSolutions = 0;

  /**
   * count of the backtracks made during search
   */
//  public int nbBacktracks = 0;

  /**
   * maximal search depth for logging statements
   */
  public int loggingMaxDepth = 5;

  /**
   * /**
   * the goal that needs be solved to drive the exploration
   */
  public AbstractIntBranching mainGoal;

  /**
   * the index of the world where the search started
   */
  public int baseWorld = 0;

  /**
   * A set of limits controlling the exploration
   */
//  public AbstractGlobalSearchLimit[] limits;
  public ArrayList limits;

  protected AbstractGlobalSearchSolver(Problem pb) {
    problem = pb;
//    limits = new AbstractGlobalSearchLimit[0];
    limits = new ArrayList();
    traceStack = new ArrayList();
    currentTraceIndex = -1;
    nextMove = INIT_SEARCH;
  }

  /**
   * main entry point: searching for one solution
   * Note: the initial propagation must be done before pushing any world level.
   * It is therefore kept before restoring a solution
   */
  public void run() {
    boolean feasibleRootState = true;
    try {
      newTreeSearch();
      problem.propagate();
    } catch (ContradictionException e) {
      feasibleRootState = false;
    }
    if (feasibleRootState) {
      problem.worldPush();
      mainGoal.explore(1);
      problem.worldPop();
      // explore is responsible for recordinf problem.feasible = Boolean.TRUE in case a solution is found
      if (problem.feasible == null) {
        problem.feasible = Boolean.TRUE;
      } else if ((maxNbSolutionStored > 0) && (!stopAtFirstSol) && existsSolution()) {
        restoreBestSolution();
      }
    } else {
      problem.feasible = Boolean.FALSE;
    }
    endTreeSearch();
  }

  /**
   * main entry point: searching for one solution
   * Note: the initial propagation must be done before pushing any world level.
   * It is therefore kept before restoring a solution
   */
  public void incrementalRun() {
    boolean feasibleRootState = true;
    try {
      newTreeSearch();
      problem.propagate();
    } catch (ContradictionException e) {
      feasibleRootState = false;
    }
    if (feasibleRootState) {
      //problem.worldPush();
      if (stopAtFirstSol) {
        nextSolution();
      } else {
        while (nextSolution() == Boolean.TRUE) {
          ;
        }
      }
      //problem.worldPop();
      //if ((maxNbSolutionStored > 0) && (!stopAtFirstSol) && existsSolution()) {
      //  restoreBestSolution();
      //}
    } else {
      problem.feasible = Boolean.FALSE;
    }
    endTreeSearch();
  }


  /**
   * called before a new search tree is explored
   */
  public void newTreeSearch() throws ContradictionException {
    assert(problem.getSolver().getSearchSolver() == this);
    for (int i = 0; i < limits.size(); i++) {
      AbstractGlobalSearchLimit lim = (AbstractGlobalSearchLimit) limits.get(i);
      lim.reset(true);
    }
    nbSolutions = 0;
    baseWorld = problem.getEnvironment().getWorldIndex();
  }

  /**
   * called before a new search tree is explored
   */
  public void endTreeSearch() {
    for (int i = 0; i < limits.size(); i++) {
      AbstractGlobalSearchLimit lim = (AbstractGlobalSearchLimit) limits.get(i);
      lim.reset(false);
    }
    if (logger.isLoggable(Level.SEVERE)) {
      if (problem.feasible == Boolean.TRUE) { // <hca> bug quand feasible est  null
        logger.log(Level.SEVERE, "solve => " + new Integer(nbSolutions) + " solutions");
      } else {
        logger.severe("solve => no solution");
      }
      for (int i = 0; i < limits.size(); i++) {
        AbstractGlobalSearchLimit lim = (AbstractGlobalSearchLimit) limits.get(i);
        logger.severe(lim.pretty());
      }
    }
  }

  /**
   * called before a node is expanded in the search tree (choice point creation)
   */
  public void newTreeNode() throws ContradictionException {
    for (int i = 0; i < limits.size(); i++) {
      AbstractGlobalSearchLimit lim = (AbstractGlobalSearchLimit) limits.get(i);
      if (lim.newNode(this) == false) {
        encounteredLimit = true;
        problem.getPropagationEngine().raiseContradiction(lim);
      }
    }
  }

  /**
   * called after a node is expanded in the search tree (choice point creation)
   */
  public void endTreeNode() throws ContradictionException {
    for (int i = 0; i < limits.size(); i++) {
      AbstractGlobalSearchLimit lim = (AbstractGlobalSearchLimit) limits.get(i);
      lim.endNode(this);
    }
  }

  /**
   * called when a solution is encountered: printing and, if needed, storing the solution
   */
  public void recordSolution() {
    problem.feasible = Boolean.TRUE;
    nbSolutions = nbSolutions + 1;
//      trace(SVIEW,"... solution [~S]\n",a.limits),  // v1.011 <thb>
    if (maxNbSolutionStored > 0) {
      super.recordSolution();
    }
  }

  /**
   * called before going down into each branch of the choice point
   */
  public void postDynamicCut() throws ContradictionException {
  }


  public IntBranchingTrace pushTrace() {
    currentTraceIndex++;
    IntBranchingTrace nextTrace = null;
    if (currentTraceIndex > traceStack.size() - 1) {
      nextTrace = new IntBranchingTrace();
      traceStack.add(nextTrace);
    } else {
      nextTrace = (IntBranchingTrace) traceStack.get(currentTraceIndex);
      nextTrace.clear();
    }
    return nextTrace;
  }

  public IntBranchingTrace popTrace() {
    if (currentTraceIndex <= 0) {
      return null;
    } else {
      currentTraceIndex--;
      return (IntBranchingTrace) traceStack.get(currentTraceIndex);
    }
  }
  public IntBranchingTrace topTrace() {
    if (currentTraceIndex < 0) {
      return null;
    } else {
      return (IntBranchingTrace) traceStack.get(currentTraceIndex);
    }
  }

  public void popTraceUntil(int targetWorld) {
    int deltaWorld = (problem.getEnvironment().getWorldIndex() - targetWorld);
    if (deltaWorld > 0) {
      currentTraceIndex = currentTraceIndex - deltaWorld;
    }
  }

  public Boolean nextSolution() {

    int previousNbSolutions = nbSolutions;
    encounteredLimit = false;
    IntBranchingTrace ctx = null;
    boolean stop = false;

    // specific initialization for the very first solution search (start from the tree root, instead of last leaf)
    if (nextMove == INIT_SEARCH) {
      nextMove = OPEN_NODE;
      ctx = new IntBranchingTrace(mainGoal);
    } else {
      ctx = topTrace();
    }
    while (!stop) {
      switch (nextMove) {
        case OPEN_NODE:
          {
            System.out.println("Open !");
            try {
              newTreeNode();
              Object branchingObj = null;
              AbstractIntBranching currentBranching = (AbstractIntBranching) ctx.getBranching();
              AbstractIntBranching nextBranching = currentBranching;
              do {
                currentBranching = nextBranching;
                branchingObj = currentBranching.selectBranchingObject();
                nextBranching = (AbstractIntBranching) currentBranching.getNextBranching();
              } while ((branchingObj == null) && (nextBranching != null));
              if (branchingObj != null) {
                ctx = pushTrace();
                ctx.setBranching(currentBranching);
                ctx.setBranchingObject(branchingObj);
                ctx.setBranchIndex(currentBranching.getFirstBranch(ctx.getBranchingObject()));
                nextMove = DOWN_BRANCH;
              } else {
                recordSolution();
                showSolution();
                nextMove = UP_BRANCH;
                stop = true;
              }
            } catch (ContradictionException e) {
              nextMove = UP_BRANCH;
            }
            break;
          }
        case UP_BRANCH:
          {
            System.out.println("Up !!" + ctx.getBranchingObject().toString() + ":" +  ctx.getBranchIndex());
            if (ctx == null) {
              stop = true;
            } else {
              try {
                problem.getEnvironment().worldPop();
                endTreeNode();
                postDynamicCut();
                ctx.getBranching().goUpBranch(ctx.getBranchingObject(), ctx.getBranchIndex());
                if (!ctx.getBranching().finishedBranching(ctx.getBranchingObject(), ctx.getBranchIndex())) {
                  ctx.setBranchIndex(ctx.getBranching().getNextBranch(ctx.getBranchingObject(), ctx.getBranchIndex()));
                  nextMove = DOWN_BRANCH;
                } else {
                  ctx = popTrace();
                  if (ctx == null) {
                    stop = true;
                    // nextMove = DONE
                  } else {
                    nextMove = UP_BRANCH;
                  }
                }
              } catch (ContradictionException e) {
                nextMove = UP_BRANCH;
              }
            }
            break;
          }
        case DOWN_BRANCH:
          {
            System.out.println("Down !!" + ctx.getBranchingObject().toString() + ":" +  ctx.getBranchIndex());
            try {
              problem.getPropagationEngine().checkCleanState();
              problem.getEnvironment().worldPush();
              ctx.getBranching().goDownBranch(ctx.getBranchingObject(), ctx.getBranchIndex());
              nextMove = OPEN_NODE;
            } catch (ContradictionException e) {
              nextMove = UP_BRANCH;
            }
            break;
          }
      }
    }
    if (nbSolutions > previousNbSolutions) {
      return Boolean.TRUE;
    } else if (encounteredLimit) {
      return null;
    } else {
      return Boolean.FALSE;
    }
  }
}
