package choco.test.real;

import junit.framework.TestCase;
import choco.real.RealInterval;
import choco.real.RealMath;
import choco.real.exp.RealIntervalConstant;

import java.util.Random;

/**
 * J-CHOCO
 * Copyright (C) F. Laburthe, 1999-2003
 * <p/>
 * An open-source Constraint Programming Kernel
 * for Research and Education
 * <p/>
 * Created by: Guillaume on 18 juin 2004
 */
public class IATests extends TestCase {
  RealInterval[] intervals = new RealInterval[3];
  public static int nbBox = 20;

  public void setUp() {
    Random rand = new Random();
    for (int i = 0; i < intervals.length; i++) {
      double a = rand.nextDouble() * 10 * (rand.nextBoolean()?-1:1);
      double b = rand.nextDouble() * 10 * (rand.nextBoolean()?-1:1);
      intervals[i] = new RealIntervalConstant(Math.min(a,b), Math.max(a,b));
    }
  }

  public void tearDown() {
    for (int i = 0; i < intervals.length; i++) {
      intervals[i] = null;
    }
  }

  public void testPlus() {
    RealInterval a = intervals[0];
    RealInterval b = intervals[1];
    System.out.println("Testing " + a + " + " + b);

    RealInterval res = RealMath.add(a,b);
    double aw = (a.getSup() - a.getInf()) / nbBox;
    double bw = (b.getSup() - b.getInf()) / nbBox;
    for(int i = 1; i < nbBox; i ++) {
      double aa = a.getInf() + aw * i;
      for(int j = 1; j < nbBox; j++) {
        double bb = b.getInf() + bw * j;
        assertTrue(aa + bb > res.getInf());
        assertTrue(aa + bb < res.getSup());
      }
    }
  }

  public void testMinus() {
    RealInterval a = intervals[0];
    RealInterval b = intervals[1];
    System.out.println("Testing " + a + " - " + b);

    RealInterval res = RealMath.sub(a,b);
    double aw = (a.getSup() - a.getInf()) / nbBox;
    double bw = (b.getSup() - b.getInf()) / nbBox;
    for(int i = 1; i < nbBox; i ++) {
      double aa = a.getInf() + aw * i;
      for(int j = 1; j < nbBox; j++) {
        double bb = b.getInf() + bw * j;
        assertTrue(aa - bb > res.getInf());
        assertTrue(aa - bb < res.getSup());
      }
    }
  }

  public void testMult() {
    RealInterval a = intervals[0];
    RealInterval b = intervals[1];
    System.out.println("Testing " + a + " * " + b);

    RealInterval res = RealMath.mul(a,b);
    double aw = (a.getSup() - a.getInf()) / nbBox;
    double bw = (b.getSup() - b.getInf()) / nbBox;
    for(int i = 1; i < nbBox; i ++) {
      double aa = a.getInf() + aw * i;
      for(int j = 1; j < nbBox; j++) {
        double bb = b.getInf() + bw * j;
        assertTrue(aa * bb > res.getInf());
        assertTrue(aa * bb < res.getSup());
      }
    }
  }

  public void testDiv() {
    RealInterval a = intervals[0];
    RealInterval b = intervals[1];
    RealInterval c = intervals[2];
    System.out.println("Testing " + a + " / " + b + " in " + c);

    RealInterval res = RealMath.odiv_wrt(a,b, c);
    double aw = (a.getSup() - a.getInf()) / nbBox;
    double bw = (b.getSup() - b.getInf()) / nbBox;
    for(int i = 1; i < nbBox; i ++) {
      double aa = a.getInf() + aw * i;
      for(int j = 1; j < nbBox; j++) {
        double bb = b.getInf() + bw * j;
        if (bb != 0) {
          assertTrue(aa / bb > res.getInf() || aa / bb > c.getSup() || aa / bb < c.getInf());
          assertTrue(aa / bb < res.getSup() || aa / bb > c.getSup() || aa / bb < c.getInf());
        }
      }
    }
  }

  public void testIPower() {
    int[] power = new int[]{2,3,4};
    RealInterval a = intervals[0];

    for (int powerIdx = 0; powerIdx < power.length; powerIdx++) {
      int p = power[powerIdx];
      System.out.println("Testing " + a + " ** " + p);
      RealInterval res = RealMath.iPower(a, p);

      double aw = (a.getSup() - a.getInf()) / nbBox;
      for(int i = 1; i < nbBox; i ++) {
        double aa = a.getInf() + aw * i;
        assertTrue(Math.pow(aa, p) > res.getInf());
        assertTrue(Math.pow(aa, p) < res.getSup());
      }
    }
  }

  /*public void testIRoot() {
    int[] power = new int[]{2,3,4};
    RealInterval a = intervals[0];
    RealInterval b = intervals[1];

    if (a.getInf() > 0) for (int powerIdx = 0; powerIdx < power.length; powerIdx++) {
      int p = power[powerIdx];
      System.out.println("Testing " + a + " ** 1/" + p);
      RealInterval res = RealMath.iRoot(a, p, b);

      double aw = (a.getSup() - a.getInf()) / nbBox;
      for(int i = 1; i < nbBox; i ++) {
        double aa = a.getInf() + aw * i;
        double calc = Math.pow(aa, 1/p);
        assertTrue(calc > res.getInf() || calc > b.getSup() || calc < b.getInf());
        assertTrue(calc < res.getSup() || calc > b.getSup() || calc < b.getInf());
      }
    }
  } */

  public void testSin() {
    RealInterval a = intervals[0];

    System.out.println("Testing sin(" + a + ")");
    RealInterval res = RealMath.sin(a);

    double aw = (a.getSup() - a.getInf()) / nbBox;
    for(int i = 1; i < nbBox; i ++) {
      double aa = a.getInf() + aw * i;
      assertTrue(Math.sin(aa) > res.getInf());
      assertTrue(Math.sin(aa) < res.getSup());
    }
  }

  public void testCos() {
    RealInterval a = intervals[0];

    System.out.println("Testing cos(" + a + ")");
    RealInterval res = RealMath.cos(a);

    double aw = (a.getSup() - a.getInf()) / nbBox;
    for(int i = 1; i < nbBox; i ++) {
      double aa = a.getInf() + aw * i;
      assertTrue(Math.cos(aa) > res.getInf());
      assertTrue(Math.cos(aa) < res.getSup());
    }
  }
}
