package chocosamples;

import choco.Problem;
import choco.integer.IntVar;

import java.util.logging.Logger;
import java.util.ArrayList;

// **************************************************
// *                   J-CHOCO                      *
// *   Copyright (C) F. Laburthe, 1999-2003         *
// **************************************************
// *  an open-source Constraint Programming Kernel  *
// *     for Research and Education                 *
// **************************************************

/* File chocosamples.U2planning.java, last modified by flaburthe 12 mars 2004 10:47:45 */


/* -- benchmark proposed by Michel Lemaitre.

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
//           ============  The U2 problem =============
//
// # "U2" has a Christmas concert that starts in 17 minutes and they must
// # all cross a bridge to get there.  All four men begin on the same side
// # of the bridge.  You must help them across to the other side. It is
// # night. There is one flashlight.  A maximum of two people can cross at
// # one time.  Any party who crosses, either 1 or 2 people, must have the
// # flashlight with them.  The flashlight must be walked back and forth,
// # it cannot be thrown, etc.  Each band member walks at a different
// # speed.  A pair must walk together at the rate of the slower man's
// # pace:
// # 
// #            Bono:- 1 minute to cross
// #            Edge:- 2 minutes to cross
// #            Adam:- 5 minutes to cross
// #            Larry:-10 minutes to cross
// # 
// # For example: if Bono and Larry walk across first, 10 minutes have
// # elapsed when they get to the other side of the bridge.  If Larry then
// # returns with the flashlight, a total of 20 minutes have passed and you
// # have failed the mission.  Notes: There is no trick behind this.  It is
// # the simple movement of resources in the appropriate order.  There are
// # two known answers to this problem.  This is based on a question
// # Microsoft gives to all prospective employees.  Note: Microsoft expects
// # you to answer this question in under 5 minutes!  Good Luck!  REMEMBER
// # - all the parameters you need to solve the problem are given to you!
// # That means all 4 guys ARE AT THE OTHER SIDE in 17 minutes. 2 guys at a
// # time MAX on the bridge at any time AND they HAVE to have the ONE
// # flashlight with them.

*/


public class U2planning {

 public static class Person {
    String name;
    int rate;

    public Person(String n, int r) {
      name = n;
      rate = r;
    }
  }
  private static Logger logger = Logger.getLogger("choco.samples.u2");
  private static int nbPersons = 5;
  private static int flashlightIndex = nbPersons - 1;
  private static int maxDurationU2 = 17;

  private static Problem pb;
  private static Person bono, edge, adam, larry, flashlight;
  private static Person[] persons;
  private static int[] timeTravel;

  public static void main(String args[]) {
    logger.fine("U2 planning problem...");
    bono = new Person("Bono", 1);
    edge = new Person("edge", 2);
    adam = new Person("adam", 5);
    larry = new Person("larry", 10);
    flashlight = new Person("flashlight", 0);
    persons = new Person[] {larry, adam, edge, bono, flashlight};
    timeTravel = new int[] {larry.rate, adam.rate, edge.rate, bono.rate, flashlight.rate};
    pb = new Problem();

    u2(12);
  }

  public static void u2(int nbSteps) {
    int nbStates = (nbSteps + 1);
    pb = new Problem();
    IntVar[][] move = pb.makeBoundIntVarArray("move", nbSteps, nbPersons, 0, 1);
    IntVar[][] position = pb.makeBoundIntVarArray("position", nbStates, nbPersons, 0, 1);
    IntVar[][] outward = pb.makeBoundIntVarArray("outward", nbSteps, nbPersons, 0, 1);
    IntVar[][] backward = pb.makeBoundIntVarArray("backward", nbSteps, nbPersons, 0, 1);
    IntVar[] duration = pb.makeBoundIntVarArray("duration", nbSteps, 1, maxDurationU2);

    for (int u=0; u<nbPersons; u++) {
      //  Starting and ending positions. 0 is before the bridge, 1 is after crossing.
      pb.post(pb.eq(position[0][u], 0));          // starting positions
      pb.post(pb.eq(position[nbStates - 1][u], 1));   // ending positions

      for (int s=0; s<nbSteps-1; s++) {
        // Define movements from positions.
        pb.post( pb.ifOnlyIf(pb.eq(outward[s][u], 1),  pb.leq(position[s][u], pb.minus(position[s + 1][u], 1))) );
        pb.post( pb.ifOnlyIf(pb.eq(backward[s][u], 1), pb.leq(position[s + 1][u], pb.minus(position[s][u], 1))) );
        pb.post( pb.eq(move[s][u], pb.plus(outward[s][u], backward[s][u])) );
        // a redundant constraint that could be useful if we were to instantiate move variables instead of position
        pb.post( pb.ifOnlyIf( pb.eq(move[s][u], 0), pb.eq(position[s + 1][u], position[s][u]) ) );

        // redundant constraint (because the cDefDuration one only propagates once all but one are instantiated)
        pb.post( pb.geq(duration[s],  pb.mult( persons[u].rate, move[s][u])) );

        // USELESS pb.post( pb.implies( pb.leq(duration[s], persons[u].rate - 1), pb.leq(move[s][u], 0)) ); //counter-opposite

        //  If somebody crosses, the flashlight also crosses in the same direction.
        pb.post( pb.leq( outward[s][u], outward[s][flashlightIndex]) );
        pb.post( pb.leq( backward[s][u], backward[s][flashlightIndex]) );
      }
    }
    for (int s=0; s<nbSteps; s++) {
      pb.post( pb.eq(move[s][flashlightIndex], 1)); //the flashlight always moves
//      pb.post( pb.eq( duration[s], pb.max( pb.scalar(move[s], timeTravel))));  //  Connect durations and moves.
//      for (int u=0; u<nbPersons; u++) {
//        pb.post( pb.leq( duration[s], pb.mult(timeTravel[u], move[s][u])));
//      }
      pb.post( pb.geq( pb.sum(move[s]), 2) );      // At each step, at least 2 (including the flashlight) are moving.
      pb.post( pb.leq( pb.sum(move[s]), 3) );      //  At each step, at most 3 (including the flashlight) are moving.

      pb.post( pb.leq(pb.sum(duration), maxDurationU2) ); // The constraint over the total duration.

      // Try to solve the problem
      pb.solve(false);
    }
  }
}
