import choco.*;
import choco.integer.*;
import choco.util.*;
import java.util.*;
import java.io.*;
import java.text.*;
import org.aspectj.lang.*;

/*
 * Created on 2 juil. 2004
 */

/**
 * Oadymppac trace implementation for Choco and Palm (DTD v2.1,
 * cf http://contraintes.inria.fr/OADymPPaC/).
 * 
 * Implemented by Guillaume Rochart (EMN/LINA)
 */
public abstract aspect  AbstractTrace {
	// ___ Event attribute configuration ___
	public static final int ATTRIBUTE_CHRONO = 0;
	public static final int ATTRIBUTE_DEPTH = 1;
	public static final int ATTRIBUTE_TIME = 2;
	public static final int ATTRIBUTE_LINE = 3;
	public static final int ATTRIBUTE_FILE = 4;
	public static final int ATTRIBUTE_NB = 5;
	public static final String[] ATTRIBUTE_NAMES = new String[]{
			"chrono", "depth", "time", "line", "file"
	};
	protected static boolean[] attributeConfig = new boolean[]{
			true, // Chrono
			true, // Depth
			false, // Time
			false, // Line
			false // File
	};
	
	// ___ Event switch configuration ___
	public static final int EVENT_NEWVAR = 0;
	public static final int EVENT_NEWCST = 1;
	public static final int EVENT_CHOICEPT = 2;
	public static final int EVENT_BACKTO = 3;
	public static final int EVENT_SOLUTION = 4;
	public static final int EVENT_FAILURE = 5;
	public static final int EVENT_REDUCE = 6;
	public static final int EVENT_RESTORE = 7;
	public static final int EVENT_EXPLAIN = 8;
	public static final int EVENT_POST = 9;
	public static final int EVENT_REMOVE = 10;
	public static final int EVENT_CONTEXT = 11;
	public static final int EVENT_AWAKE = 12;
	public static final int EVENT_NB = 13;
	public static final String[] EVENT_NAMES = new String[]{
			"newvar", "newcst", "choicept", "backto", "solution", "failure",
			"reduce", "restore", "explain", "post", "remove", "context", "awake"
	};
	protected static boolean[] eventConfig = new boolean[]{
			true, true, true, true, true, true,
			true, false, false, true, false, false, true
	};
	
	// Trace output inf a file ?
	protected boolean outputInFile = false;
	
	// Configuration of the tracer
	protected Properties traceConfig = new Properties();
	
	// Print stream where Trace must be generated
	protected PrintStream stream;
	
	// Current indentation level
	protected int indent;
	// PreAffected indentation spaces
	protected final String[] indents = new String[]{"", " ", "  ", "   ", "    ",
			"     ", "      ", "       ", "        "};
	
	// Program name
	protected String programName;
	
	// Current problem
	protected Problem currentProblem;
	
	// Trace execution information
	protected int varNb = 0;
	protected int constNb = 0;
	protected int eventNb = 0;
	protected long startTime = System.currentTimeMillis();
	
	// Inter type modifications
	public int IntVar.varNumber = -1;
	public int Constraint.constNumber = -1;
	
	
	// ==============================================
	// |                                            |
	// |                 PointCuts                  |
	// |                                            |
	// ==============================================
	
	// Program execution
	pointcut programExecution() : execution(public static void main(String[])) &&
		!cflowbelow(execution(public static void main(String[])));
	
	// Problem creation
	pointcut problemCreation() : call(Problem+.new(..));
	
	// Variable creation
	pointcut variableCreation() : call(IntVar+.new(..)) &&
		!cflow(execution(* Problem+.solve*(..)));
	
	// User constraint creation
	pointcut constraintCreation() : call(IntConstraint+.new(..)) && 
		!cflow(execution(* Problem+.solve*(..)));
	
	// Classical constraint post
	pointcut constraintPost(IntConstraint c) : execution(void Problem+.post(Constraint)) &&
		args(c) && !cflow(execution(* Problem+.solve*(..)));
	
	// Solution storing
	pointcut solution(AbstractSolver solver) : 
		execution(void AbstractSolver+.recordSolution()) &&
		!cflowbelow(execution(void AbstractSolver+.recordSolution())) &&
		target(solver);
	
	// Any constraint awake
	pointcut anyAwake() : execution(* IntConstraint+.awake*(..)) ||
		execution(* IntConstraint+.propagate(..));
	// Awake on lower bound
	pointcut awakeOnInf(IntConstraint c, int idx) :
		!cflowbelow(anyAwake()) && 
		execution(* IntConstraint+.awakeOnInf(..)) &&
		target(c) && args(idx);
	// Awake on upper bound
	pointcut awakeOnSup(IntConstraint c, int idx) :
		!cflowbelow(anyAwake()) &&
		execution(* IntConstraint+.awakeOnSup(..)) && 
		target(c) && args(idx);
	// Awake on both bounds
	pointcut awakeOnBounds(IntConstraint c, int idx) :
		!cflowbelow(anyAwake()) &&
		execution(* IntConstraint+.awakeOnBounds(..)) &&
		target(c) && args(idx);
	// Awake on instantiation
	pointcut awakeOnInst(IntConstraint c, int idx) :
		!cflowbelow(anyAwake()) &&
		execution(* IntConstraint+.awakeOnInst(..)) &&
		target(c) && args(idx);
	// Awake on removals
	pointcut awakeOnRemovals(IntConstraint c, int idx) :
		!cflowbelow(anyAwake()) &&
		execution(* IntConstraint+.awakeOnRemovals(..)) &&
		target(c) && args(idx, *);
	// Awake or propagate
	pointcut awakeOrPropagate(IntConstraint c) :
		!cflowbelow(anyAwake()) && target(c) &&
		(execution(* IntConstraint+.awake(..)) || execution(* IntConstraint+.propagate(..)));
	
	// ==============================================
	// |                                            |
	// |                  Advices                   |
	// |                                            |
	// ==============================================
	
	// Program start
	before() : programExecution() {
		init(thisJoinPointStaticPart);
		stream.println("<!DOCTYPE gentra4cp SYSTEM " +
				"\"http://contraintes.inria.fr/OADymPPaC/Public/Trace/gentra4cp.2.1.dtd\">");
		elementOpening("<gentra4cp>");
		header();
		provide();
	}
	
	// Program end
	after() : programExecution() {
		elementClosing("</gentra4cp>");
		stream.flush();
	}
	
	// Problem creation
	after() returning(Problem pb): problemCreation() {
		currentProblem = pb;
	}
	
	// Variable creation   *********** NEW VAR ***********
	after() returning(IntVar var): variableCreation() {
		var.varNumber = varNb++;
		elementOpening("<new-variable " + eventAttributes(thisJoinPointStaticPart) +
				" vident=\"v" + var.varNumber + "\" vname=\"" + var + "\">");
		elementOpening("<vardomain min=\"" + var.getInf() + "\" max=\"" + 
				var.getSup() + "\" size=\"" + var.getDomainSize() + "\">");
		if (!var.hasEnumeratedDomain() || var.getSup() - var.getInf() + 1 == 
			var.getDomainSize()) {
			singleElementLn("<range from=\"" + var.getInf() + "\" to=\"" + var.getSup() + "\"/>");
		} else {
			StringBuffer str = new StringBuffer();
			str.append("<values>");
			for(IntIterator iter = var.getDomain().getIterator(); iter.hasNext();) {
				int val = iter.next();
				str.append(val + " ");
			}
			str.append("</values>");
			singleElementLn(str.toString());
		}
		elementClosing("</vardomain>");
		elementClosing("</new-variable>");
	}
	
	// Constraint creation   *********** NEW CST ***********
	after() returning(IntConstraint c): constraintCreation() {
		c.constNumber = constNb++;
		singleElementLn("<new-constraint " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + c.constNumber + "\" cexternal=\"" + c + "\" orig=\"user\"/>");
	}
	
	// Constraint post   *********** POST ***********
	after(IntConstraint c) : constraintPost(c) {
		singleElementLn("<post " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + c.constNumber + "\"/>");
	}
	
	// Solution storing   *********** SOLUTION ***********
	after(AbstractSolver solver) : solution(solver) {
		Problem pb = solver.problem;
		elementOpening("<solution " + eventAttributes(thisJoinPointStaticPart) + ">");
		elementOpening("<state>");
		for(int v = 0; v < pb.getNbIntVars(); v++) {
			singleElementLn("<variable vident=\"v" + pb.getIntVar(v).varNumber +
					"\"><vardomain><values>" + pb.getIntVar(v).getValue() +
					"</values></vardomain></variable>");
		}
		elementClosing("</state>");
		elementClosing("</solution>");
	}
	
	//   *********** AWAKE ***********
	// Awake on lower bound   
	before(IntConstraint cst, int idx) : awakeOnInf(cst, idx) {
		elementOpening("<awake " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + cst.constNumber + "\">");
		singleElementLn("<update vident=\"v" + ((IntVar)cst.getVar(idx)).varNumber + 
				"\" types=\"min\"/>");
		elementClosing("</awake>");
	}
	
	// Awake on upper bound
	before(IntConstraint cst, int idx) : awakeOnSup(cst, idx) {
		elementOpening("<awake " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + cst.constNumber + "\">");
		singleElementLn("<update vident=\"v" + ((IntVar)cst.getVar(idx)).varNumber + 
				"\" types=\"max\"/>");
		elementClosing("</awake>");
	}
	
	// Awake on both bounds
	before(IntConstraint cst, int idx) : awakeOnBounds(cst, idx) {
		elementOpening("<awake " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + cst.constNumber + "\">");
		singleElementLn("<update vident=\"v" + ((IntVar)cst.getVar(idx)).varNumber + 
				"\" types=\"minmax\"/>");
		elementClosing("</awake>");
	}
	
	// Awake on instantiation
	before(IntConstraint cst, int idx) : awakeOnInst(cst, idx) {
		elementOpening("<awake " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + cst.constNumber + "\">");
		singleElementLn("<update vident=\"v" + ((IntVar)cst.getVar(idx)).varNumber + 
				"\" types=\"ground\"/>");
		elementClosing("</awake>");
	}
	
	// Awake on removals
	before(IntConstraint cst, int idx) : awakeOnRemovals(cst, idx) {
		elementOpening("<awake " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + cst.constNumber + "\">");
		singleElementLn("<update vident=\"v" + ((IntVar)cst.getVar(idx)).varNumber + 
				"\" types=\"val\"/>");
		elementClosing("</awake>");
	}
	
	// Awake on constraint
	before(IntConstraint cst) : awakeOrPropagate(cst) {
		elementOpening("<awake " + eventAttributes(thisJoinPointStaticPart) +
				" cident=\"c" + cst.constNumber + "\">");
		elementClosing("</awake>");
	}
	
	// ==============================================
	// |                                            |
	// |                    Tools                   |
	// |                                            |
	// ==============================================
	
	// Stream printing
	public void elementOpening(String str) {
		int realInd = Math.min(indent, indents.length - 1);
		stream.println(indents[realInd] + str);
		indent++;
	}
	
	public void singleElementLn(String str) {
		int realInd = Math.min(indent, indents.length - 1);
		stream.println(indents[realInd] + str);
	}
	
	public void singleElement(String str) {
		int realInd = Math.min(indent, indents.length - 1);
		stream.print(indents[realInd] + str);
	}
	
	public void elementClosing(String str) {
		indent--;
		int realInd = Math.min(indent, indents.length - 1);
		stream.println(indents[realInd] + str);
	}
	
	// Initialization
	protected void init(JoinPoint.StaticPart jp) {
		String value;
		// Configuration update
		try {
			traceConfig.load(new FileInputStream("trace.properties"));
		} catch(IOException e) {}
		for(int i = 0; i < ATTRIBUTE_NB; i++) {
			value = (String) traceConfig.get("attribute." + ATTRIBUTE_NAMES[i]);
			if (value != null) {
				attributeConfig[i] = new Boolean(value).booleanValue();
			}
		}
		value = (String) traceConfig.get("output");
		if (value.equals("file")) outputInFile = true;
		for(int i = 0; i < EVENT_NB; i++) {
			value = (String) traceConfig.get("event." + EVENT_NAMES[i]);
			if (value != null) {
				eventConfig[i] = new Boolean(value).booleanValue();
			}
		}
		
		// Program name
		programName = jp.getSourceLocation().getFileName();
		
		// Stream creation
		if (outputInFile) {
			try {
				File file = 
					new File(programName.substring(0, programName.length() - 5) 
					+ "-" + solverName() + ".xml");
				OutputStream tmp = new FileOutputStream(file);
				OutputStream tmp2 = new BufferedOutputStream(tmp);
				stream = new PrintStream(tmp2);
			} catch(Exception e) {
				System.err.println("File creation failed -> display on default output.");
				stream = System.out;
			}
		} else {
			stream = System.out;
		}
		
	}
	
	// Header
	protected void header() {
		elementOpening("<header>");
		Date today = new Date();
		DateFormat format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		singleElementLn("<date>" + format.format(today) + "</date>");
		singleElementLn("<source>" + programName + "</source>");
		singleElementLn("<contributor>Guillaume Rochart -- EMN/LINA</contributor>");
		singleElementLn("<rights>BSD license</rights>");
		singleElementLn("<solver>" + solverName() + "</solver>");
		elementClosing("</header>");
	}
	
	// Provide
	protected void provide() {
		String eventAttributes = "" + 
			(attributeConfig[ATTRIBUTE_CHRONO]?"chrono=\"\" ":"") +
			(attributeConfig[ATTRIBUTE_DEPTH]?"depth=\"\" ":"") +
			(attributeConfig[ATTRIBUTE_TIME]?"time=\"\" ":"") +
			(attributeConfig[ATTRIBUTE_LINE]?"line=\"\" ":"") +
			(attributeConfig[ATTRIBUTE_FILE]?"file=\"\" ":"");
		elementOpening("<provide>");
		if (eventConfig[EVENT_NEWVAR]) {
			elementOpening("<new-variable " + eventAttributes + "vident=\"\" vname=\"\">");
			elementOpening("<vardomain min=\"\" max=\"\" size=\"\">");
			singleElementLn("<values/> <range from=\"\" to=\"\"/>");
			elementClosing("</vardomain>");
			elementClosing("</new-variable>");
		}
		if (eventConfig[EVENT_NEWCST]) {
			singleElementLn("<new-constraint " + eventAttributes + "cident=\"\" cexternal=\"\" orig=\"\"/>");
		}
		if (eventConfig[EVENT_CHOICEPT]) {
			singleElementLn("<choice-point " + eventAttributes + "/>");
		}
		if (eventConfig[EVENT_BACKTO]) {
			singleElementLn("<back-to " + eventAttributes + "/>");
		}
		if (eventConfig[EVENT_SOLUTION]) {
			elementOpening("<solution " + eventAttributes + ">");
			elementOpening("<state>");
			singleElementLn("<variable vident=\"\"><vardomain><values/></vardomain></variable>");
			elementClosing("</state>");
			elementClosing("</solution>");
		}
		if (eventConfig[EVENT_FAILURE]) {
			singleElementLn("<failure " + eventAttributes + "/>");
		}
		if (eventConfig[EVENT_AWAKE]) {
			elementOpening("<awake " + eventAttributes + "cident=\"\">");
			singleElementLn("<update vident=\"\" types=\"\"/>");
			elementClosing("</awake>");
		}
		if (eventConfig[EVENT_REDUCE]) {
			elementOpening("<reduce " + eventAttributes + "cident=\"\" vident=\"\">");
			singleElementLn("<delta><values/><range from=\"\" to=\"\"/></delta>");
			singleElementLn("<update vident=\"\" types=\"\"/>");
			if (eventConfig[EVENT_EXPLAIN]) {
				elementOpening("<explanation>");
				singleElementLn("<range from=\"\" to=\"\"/>");
				singleElementLn("<constraints cidents=\"\"/>");
				elementClosing("</explanation>");
			}
			elementClosing("</reduce>");
		}
		if (eventConfig[EVENT_RESTORE]) {
			elementOpening("<restore " + eventAttributes + ">");
			singleElementLn("<delta><values/><range from=\"\" to=\"\"/></delta>");
			elementClosing("</restore>");
		}
		if (eventConfig[EVENT_POST]) {
			singleElementLn("<post " + eventAttributes + "cident=\"\"" +
					(eventConfig[EVENT_CONTEXT]?" context=\"\"":"") + "/>");
		}
		if (eventConfig[EVENT_REMOVE]) {
			singleElementLn("<remove " + eventAttributes + "cident=\"\"" +
					(eventConfig[EVENT_CONTEXT]?" context=\"\"":"") + "/>");
		}
		elementClosing("</provide>");
	}
	
	// All event attributes w.r.t. attribute configuration
	public String eventAttributes(JoinPoint.StaticPart jp) {
		StringBuffer str = new StringBuffer();
		if (attributeConfig[ATTRIBUTE_CHRONO]) str.append("chrono=\"" + (eventNb++) + "\" ");
		if (attributeConfig[ATTRIBUTE_DEPTH]) str.append("depth=\"" + getCurrentDepth() + "\" ");
		if (attributeConfig[ATTRIBUTE_TIME]) str.append("time=\"" + (System.currentTimeMillis() - startTime) + "\" ");
		if (attributeConfig[ATTRIBUTE_LINE]) str.append("line=\"" + jp.getSourceLocation().getLine() + "\" ");
		if (attributeConfig[ATTRIBUTE_FILE]) str.append("file=\"" + jp.getSourceLocation().getFileName() + "\" ");
		return str.toString();
	}
	
	protected abstract String solverName();
	protected abstract int getCurrentDepth();
}