import choco.mem.*;
import choco.*;
import choco.prop.*;
import choco.util.*;
import choco.integer.*;
import org.aspectj.lang.*;

/*
 * Created on 2 juil. 2004
 */

/**
 * Oadymppac trace implementation for Choco (DTD v2.1,
 * cf http://contraintes.inria.fr/OADymPPaC/).
 * 
 * Implemented by Guillaume Rochart (EMN/LINA)
 */
public aspect ChocoTrace extends AbstractTrace {
	// Previous domain
	protected int[]  previousDomain;
	
	// ==============================================
	// |                                            |
	// |                 PointCuts                  |
	// |                                            |
	// ==============================================
	
	// Choice point on world push
	pointcut choicePoint() : execution(public void Environment.worldPush());
	
	// Back track on world pop
	pointcut backTo() : execution(public void Environment.worldPop());
	
	// Contradiction
	pointcut failure() : call(ContradictionException.new(..)) && !cflow(anyUpdate(IntVar));
	
	// Any modification
	pointcut anyUpdate(IntVar var) : target(var) && (execution(* IntVar+.update*(..)) || 
		execution(* IntVar+.remove*(..)) || execution(* IntVar+.instantiate(..)));
	
	// Instantiation
	pointcut instantiate(Constraint cst, IntVar var):  
		execution(* ChocEngine+.postInstInt(..)) && args(var, *) &&
		cflow(anyAwake() && target(cst));
	pointcut instantiateC(IntVar var):
		execution(* ChocEngine+.postInstInt(..)) && args(var, *) &&
		!cflow(anyAwake());
	
	// Bound updates
	pointcut updateInf(Constraint cst, IntVar var):
		execution(* ChocEngine+.postUpdateInf(..)) && args(var, *) &&
		cflow(anyAwake() && target(cst));
	pointcut updateSup(Constraint cst, IntVar var):
		execution(* ChocEngine+.postUpdateSup(..)) && args(var, *) &&
		cflow(anyAwake() && target(cst));
	pointcut updateInfC(IntVar var):
		execution(* ChocEngine+.postUpdateInf(..)) && args(var, *) &&
		!cflow(anyAwake());
	pointcut updateSupC(IntVar var):
		execution(* ChocEngine+.postUpdateSup(..)) && args(var, *) &&
		!cflow(anyAwake());
	
	// Remove Val
	pointcut removeVal(Constraint cst, IntVar var, int val):
		execution(* ChocEngine+.postRemoveVal(..)) && args(var, val, *) &&
		cflow(anyAwake() && target(cst));
	pointcut removeValC(IntVar var, int val):
		execution(* ChocEngine+.postRemoveVal(..)) && args(var, val, *) &&
		!cflow(anyAwake());
	
	// ==============================================
	// |                                            |
	// |                  Advices                   |
	// |                                            |
	// ==============================================
	
	// Choice point and back track
	before() : choicePoint() {
		singleElementLn("<choice-point " + eventAttributes(thisJoinPointStaticPart) + "/>");
	}
	
	after() : backTo() {
		singleElementLn("<back-to " + eventAttributes(thisJoinPointStaticPart) + "/>");
	}
	
	// Failures (except empty domains)
	before() : failure() {
		singleElementLn("<failure " + eventAttributes(thisJoinPointStaticPart) + "/>");
	}
	
	// Stores domain before update
	before(IntVar var) : anyUpdate(var) {
		if (var.hasEnumeratedDomain()) {
			previousDomain = new int[var.getDomainSize()];
			int i = 0;
			for(IntIterator iterator = var.getDomain().getIterator(); iterator.hasNext();) {
				previousDomain[i++] = iterator.next();
			}
		} else {
			previousDomain = new int[]{var.getInf(), var.getSup()};
		}
	}
	
	// Empty domains
	after(IntVar var) throwing(ContradictionException e): anyUpdate(var) {
		elementOpening("<reduce " + eventAttributes(thisJoinPointStaticPart) + 
				" vident=\"v" + var.varNumber + "\">");
		elementOpening("<delta>");
		if (var.hasEnumeratedDomain()) {
			singleElement("<values>");
			for(int i = 0; i < previousDomain.length; i++) {
				stream.print(previousDomain[i] + " ");
			}
			stream.println("</values>");
		} else {
			singleElementLn("<range from=\"" + previousDomain[0] + "\" to=\"" +
					previousDomain[1] + "\"/>");
		}
		elementClosing("</delta>");
		singleElementLn("<update vident=\"v" + var.varNumber + "\" typees=\"empty\"/>");
		elementClosing("</reduce>");
		singleElementLn("<failure " + eventAttributes(thisJoinPointStaticPart) + "/>");
	}
	
	// Instantiation
	before(Constraint cst, IntVar var): instantiate(cst, var) {
		instantiateHandle(cst, var, thisJoinPointStaticPart);
	}
	before(IntVar var): instantiateC(var) {
		instantiateHandle(null, var, thisJoinPointStaticPart);
	}
	
	// Bound updates
	before(Constraint cst, IntVar var): updateInf(cst, var) {
		boundHandle(cst, var, 0, thisJoinPointStaticPart);
	}
	before(Constraint cst, IntVar var): updateSup(cst, var) {
		boundHandle(cst, var, 1, thisJoinPointStaticPart);
	}
	before(IntVar var): updateInfC(var) {  // Really useful ?????
		boundHandle(null, var, 0, thisJoinPointStaticPart);
	}
	before(IntVar var): updateSupC(var) {  // Really useful ?????
		boundHandle(null, var, 1, thisJoinPointStaticPart);
	}
	
	// Value removals
	before(Constraint cst, IntVar var, int val): removeVal(cst, var, val) {
		removalHandle(cst, var, val, thisJoinPointStaticPart);
	}
	before(IntVar var, int val): removeValC(var, val) {  // Really useful ?????
		removalHandle(null, var, val, thisJoinPointStaticPart);
	}
	
	// ==============================================
	// |                                            |
	// |                    Tools                   |
	// |                                            |
	// ==============================================
	
	protected String solverName() {
		return "Choco";
	}
	
	protected int getCurrentDepth() {
		return currentProblem.getWorldIndex();
	}
	
	protected void instantiateHandle(Constraint cst, IntVar var, JoinPoint.StaticPart jp) {
		elementOpening("<reduce " + eventAttributes(jp) +
				(cst != null?" cident=\"c" + cst.constNumber + "\"":"") + 
				 " vident=\"v" + var.varNumber + "\">");
		elementOpening("<delta>");
		if (var.hasEnumeratedDomain()) {
			singleElement("<values>");
			for(int i = 0; i < previousDomain.length; i++) {
				if (previousDomain[i] != var.getValue())
					stream.print(previousDomain[i] + " ");
			}
			stream.println("</values>");
		} else {
			if (previousDomain[0] < var.getInf())
				singleElementLn("<range from=\"" + previousDomain[0] + "\" to=\"" +
					(var.getInf() - 1) + "\"/>");
			if (previousDomain[1] > var.getSup())
				singleElementLn("<range from=\"" + (var.getSup() + 1) + "\" to=\"" +
				    previousDomain[1] + "\"/>");
		}
		elementClosing("</delta>");
		singleElementLn("<update vident=\"v" + var.varNumber + "\" types=\"ground\"/>");
		elementClosing("</reduce>");
	}
	
	protected void boundHandle(Constraint cst, IntVar var, int bound, JoinPoint.StaticPart jp) {
		elementOpening("<reduce " + eventAttributes(jp) +
				(cst != null?" cident=\"c" + cst.constNumber + "\"":"") + 
				 " vident=\"v" + var.varNumber + "\">");
		elementOpening("<delta>");
		if (var.hasEnumeratedDomain()) {
			singleElement("<values>");
			if (bound == 0) {
				int i = 0;
				while (i < previousDomain.length && previousDomain[i] < var.getInf()) {
					stream.print(previousDomain[i] + " ");
					i++;
				}
			} else {
				int i = 0;
				while (i < previousDomain.length && previousDomain[i] <= var.getSup()) {
					i++;
				}
				while (i < previousDomain.length) {
					stream.print(previousDomain[i] + " ");
					i++;
				}
			}
			stream.println("</values>");
		} else {
			if (bound == 0) {
				singleElementLn("<range from=\"" + previousDomain[0] + "\" to=\"" +
						(var.getInf() - 1) + "\"/>");
			} else {
				singleElementLn("<range from=\"" + (var.getSup() + 1) + "\" to=\"" +
						previousDomain[1] + "\"/>");
			}
		}
		elementClosing("</delta>");
		singleElementLn("<update vident=\"v" + var.varNumber + 
				"\" typees=\"" + (bound == 0?"min":"max") + "\"/>");
		elementClosing("</reduce>");
	}
	
	protected void removalHandle(Constraint cst, IntVar var, int val, JoinPoint.StaticPart jp) {
		elementOpening("<reduce " + eventAttributes(jp) + 
			(cst != null?" cident=\"c" + cst.constNumber + "\"":"") + ">");
		singleElementLn("<delta><values>"+val+"</values></delta>");
		singleElementLn("<update  vident=\"v" + var.varNumber + "\" types=\"val\"/>");
		elementClosing("</reduce>");
		stream.flush();
	}
}
